/**
 * @file err_lookup.cpp
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "err_lookup.h"

#include <fstream>
#include "err_codes.h"
#include "err_common.h"

// global error messages table
ErrorLookup gErrorTable;



ErrorLookup::ErrorLookup()
{
	ResHandle = NULL;
}

void ErrorLookup::Init(HMODULE ModuleHandle)
{
	ResHandle = ModuleHandle;
	Add(ERAERROR_SUCCESS, IDS_TXPLN_ERAERROR_SUCCESS, 0);
	Add(ERAERROR_UNKNOWN, IDS_TXPLN_ERAERROR_UNKNOWN, IDS_TXFMT_ERAERROR_UNKNOWN);
}

void ErrorLookup::Add(DWORD ErrorCode, UINT IdsBasic, UINT IdsFormatted)
{
	std::wstring txBasic, txFormatted;

	ASSERT(IdsBasic != 0); // musn't be zero

	LoadResStr(IdsBasic, txBasic);
	ASSERT(!txBasic.empty()); // such IDS probably doesn't exist

	if (IdsFormatted != 0) // can be zero
	{
		LoadResStr(IdsFormatted, txFormatted);
		ASSERT(!txFormatted.empty()); // such IDS probably doesn't exist
	}

	ErrorTable.push_back(ErrorMessage(ErrorCode, txBasic, txFormatted));
}


ErrorMessage *ErrorLookup::FindErrorCode(DWORD ErrorCode)
{
	for (size_t i = 0; i < ErrorTable.size(); ++i)
	{
		ErrorMessage &item = ErrorTable[i];
		if (item.Code == ErrorCode)
			return &item;
	}
	return NULL;
}


void ErrorLookup::LoadResStr(UINT ID, std::wstring &Str)
{
	PWCHAR buf;
	int size = LoadStringW(ResHandle,ID,(PWCHAR)&buf, 0);
	Str.assign(buf, size);
}


#if _DEBUG
void ErrorLookup::DumpToFile(LPCSTR FileName)
{
	std::wofstream out(FileName, std::ofstream::binary);
	for (size_t i = 0; i < ErrorTable.size(); ++i)
	{
		ErrorMessage &item = ErrorTable[i];
		out				 << L"[" << (int)item.Code << L"]" << std::endl;
		out << std::endl << L"\"" << item.BasicText << L"\"" << std::endl;
		out << std::endl << L"\"" << item.FormattedText << L"\"" << std::endl;
		out << L"------------------------------------------------" << std::endl;
	}
	out.close();
}
#endif


std::wstring GetErrorMessage(DWORD ErrorCode)
{
	ErrorMessage *found = gErrorTable.FindErrorCode(ErrorCode);
	if (found == NULL)
	{
		found = gErrorTable.FindErrorCode(ERAERROR_UNKNOWN);
		if (found == NULL)
			return L"?"; // this should never happen
		return MakeErrorMessage(found->FormattedText, std::to_wstring((unsigned long long)ErrorCode).c_str());
	}
	return found->BasicText;
}


std::wstring GetErrorMessage(DWORD ErrorCode, LPCWSTR Param1, LPCWSTR Param2, LPCWSTR Param3, LPCWSTR Param4, LPCWSTR Param5)
{
	ErrorMessage *found = gErrorTable.FindErrorCode(ErrorCode);
	if (found == NULL)
		return GetErrorMessage(ERAERROR_UNKNOWN);
	if (Param1 == NULL)
		return found->BasicText;
	return MakeErrorMessage(found->FormattedText, Param1, Param2, Param3, Param4, Param5);
}


void StrReplaceAll(std::wstring &Text, const std::wstring OldStr, const std::wstring NewStr)
{
	size_t oldLen = OldStr.length();
	if (oldLen == 0)
		return;
	size_t newLen = NewStr.length();
	size_t pos = 0;
	while (true)
	{
		pos = Text.find(OldStr, pos);
		if (pos == std::string::npos)
			return;
		Text.replace(pos, oldLen, NewStr);
		pos += newLen;
	}
}


std::wstring MakeErrorMessage(std::wstring Message, LPCWSTR Param1, LPCWSTR Param2, LPCWSTR Param3, LPCWSTR Param4, LPCWSTR Param5)
{
	if (Param1 != NULL)
		StrReplaceAll(Message, L"[1]", Param1);
	if (Param2 != NULL)
		StrReplaceAll(Message, L"[2]", Param2);
	if (Param3 != NULL)
		StrReplaceAll(Message, L"[3]", Param3);
	if (Param4 != NULL)
		StrReplaceAll(Message, L"[4]", Param4);
	if (Param5 != NULL)
		StrReplaceAll(Message, L"[5]", Param5);
	return Message;
}

