/**
 * @file CommandListParser.cpp
 * @author Copyright (c) 2014 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "CommandListParser.h"

#include <Common/era_api/err_codes.h>
#include "EracmdExceptions.h"

CommandListParser::CommandListParser(const std::wstring &input)
:
	input(input),
	comma(L';'),
	endOfLine('\n'),
	quote(L'"'),
	commentBegin(L'#'),
	fileReplacementMark(L'@')
{
	std::wostringstream s;
	s << comma << endOfLine << quote << commentBegin << fileReplacementMark;
	importantChars = s.str();
}

StringVector CommandListParser::GetCommands()
{
	commands.clear();
	currentCommand.str(L"");
	inQuotes = false;
	inComment = false;
	inFileReplacement = false;

	ResetPosition();

	while (! IsAtEnd())
	{
		currentPartBegin = pos;

		GoToNextImportantChar();

		if (! inComment)
			AppendRemainingInput();

		if (! IsAtEnd())
		{
			if (inComment)
			{
				if (IsAtLineEnd())
				{
					AddCurrentCommandToResult();
					inComment = false;
				}
			}
			else
			{
				if (IsAtCommandEnd() || IsAtLineEnd())
				{
					if (inQuotes || inFileReplacement)
						AppendCurrentChar();
					else
						AddCurrentCommandToResult();
				}

				if (IsAtQuote())
				{
					AppendCurrentChar();

					inQuotes = !inQuotes;
				}

				if (IsAtCommentBegin())
				{
					if (inQuotes || inFileReplacement)
						AppendCurrentChar();
					else
						inComment = true;
				}

				if (IsAtFileReplacementMark())
				{
					AppendCurrentChar();

					if (! inQuotes)
						inFileReplacement = !inFileReplacement;
				}
			}
		}

		Advance();
	}

	AddCurrentCommandToResult();

	if (inQuotes)
		throw EraCMDException(CMDERROR_MISSING_END_QUOTE);

	if (inFileReplacement)
		throw EraCMDException(CMDERROR_FILE_REPLACING_MISSING_END);

	return std::move(commands);
}

void CommandListParser::ResetPosition()
{
	if (input.size() > 0)
	{
		pos = 0;
	}
	else
	{
		pos = std::wstring::npos;
	}
}

void CommandListParser::GoToNextImportantChar()
{
	pos = input.find_first_of(importantChars, pos);
}

void CommandListParser::Advance()
{
	if (pos < (input.size() - 1))
	{
		++pos;
	}
	else
	{
		pos = std::wstring::npos;
	}
}

void CommandListParser::AppendRemainingInput()
{
	currentCommand << input.substr(currentPartBegin, pos - currentPartBegin);
}

void CommandListParser::AppendCurrentChar()
{
	currentCommand << input[pos];
}

void CommandListParser::AddCurrentCommandToResult()
{
	std::wstring command = currentCommand.str();
	currentCommand.str(L"");

	size_t firstValidChar = command.find_first_not_of(L" \t\n\r");
	if (firstValidChar == std::wstring::npos)
	{
		return;
	}

	size_t lastValidChar = command.find_last_not_of(L" \t\n\r");

	commands.push_back(command.substr(firstValidChar, lastValidChar + 1 - firstValidChar));
}

bool CommandListParser::IsAtEnd() const
{
	return (pos == std::wstring::npos);
}

bool CommandListParser::IsAtCommandEnd() const
{
	return (input[pos] == comma);
}

bool CommandListParser::IsAtCommentBegin() const
{
	return (input[pos] == commentBegin);
}

bool CommandListParser::IsAtQuote() const
{
	return (input[pos] == quote);
}

bool CommandListParser::IsAtFileReplacementMark() const
{
	return (input[pos] == fileReplacementMark);
}

bool CommandListParser::IsAtLineEnd() const
{
	return (input[pos] == endOfLine);
}



