/**
 * @file ScriptInterpreter.cpp
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"

#include <Common/era_api/era_api.h>
#include <Common/era_api/err_codes.h>
#include <Common/scope_exit.h>
#include "commands/_all.h"
#include "flags/CommandFlagsStack.h"
#include "flags/FlagsRegistry.h"
#include "output_format/OutputHandler.h"
#include "output_format/TextWidth_Console.h"
#include "output_format/TextWidth_File.h"
#include "util/stringUtils.h"
#include "Command.h"
#include "CommandParser.h"
#include "CommandListParser.h"
#include "ConnectionParams.h"
#include "ConsoleKeyboardInput.h"
#include "Element.h"
#include "EracmdExceptions.h"
#include "FilesReplace.h"
#include "resource.h"
#include "ScriptInterpreter.h"
#include "Token.h"

#ifdef _DEBUG
#include "debugUtils.h"
#endif

std::wstring ScriptInterpreter::lastCommand;
extern ConsoleColors_Win gConsoleColors;
extern TextWidth_Console gTextWidth_Console;
extern TextWidth_File gTextWidth_File;
extern FlagsRegistry gFlagsRegistry;

//////////////////////////////////////////////////////////////////////////
// ScriptInterpreter
//////////////////////////////////////////////////////////////////////////

ScriptInterpreter::ScriptInterpreter(CommandFlagsStack &commandFlagsStack, bool &partiallyFailed, const std::wstring &scriptPath, int recursionDepth)
:
	commandFlagsStack(commandFlagsStack), partiallyFailed(partiallyFailed), recursionDepth(recursionDepth), connectedByThis(false), shouldExit(false)
{
	commandFlagsStack.Push();

	std::shared_ptr<ParamValidatorTables> tableNameValidator = std::make_shared<ParamValidatorTables>(false);
	std::shared_ptr<ParamValidatorColumns> columnForTable0Validator = std::make_shared<ParamValidatorColumns>(0, true, false);
	std::shared_ptr<ParamValidatorColumns> columnForTable0Validator_selectableOnly = std::make_shared<ParamValidatorColumns>(0, true, true);
	std::shared_ptr<ParamValidatorColumnsWithValues> columnsWithValuesValidator = std::make_shared<ParamValidatorColumnsWithValues>(0, true);
	std::shared_ptr<ParamValidatorNoList> noListValidator = std::make_shared<ParamValidatorNoList>();
	std::shared_ptr<ParamValidatorServerCfgItems> serverCfgItemsValidator = std::make_shared<ParamValidatorServerCfgItems>(false, false);
	std::shared_ptr<ParamValidatorServerCfgItems> serverCfgPwdItemsValidator = std::make_shared<ParamValidatorServerCfgItems>(false, true);

	commandList.Add(
		L"set",
		std::shared_ptr<Command>((new Command_Set(commandFlagsStack.CurrentFlags()))
			->AddParam(0, Command::optional, IDS_HELP_SET_ARG1, std::make_shared<ParamValidatorFlag>(false, true))
			->AddParam(1, Command::optional, IDS_HELP_SET_ARG2, noListValidator))
		);

	commandList.Add(
		L"version",
		std::shared_ptr<Command>((new Command_Version())
			->AddParam(0, Command::optional, IDS_HELP_VERSION_ARG1, std::make_shared<ParamValidatorEnum>(L"api,cmd,server", false)))
		);

	commandList.Add(
		L"script",
		std::shared_ptr<Command>((new Command_RunScript(commandFlagsStack, partiallyFailed, recursionDepth))
			->AddParam(0, Command::mandatory, IDS_HELP_SCRIPT_ARG1, noListValidator))
		);

	commandList.Add(
		L"exit",
		std::shared_ptr<Command>((new Command_Exit(*this)))
		);

	commandList.Add(
		L"cls",
		std::shared_ptr<Command>((new Command_Cls()))
		);

	commandList.Add(
		L"echo",
		std::shared_ptr<Command>((new Command_Echo())
			->AddParam(0, Command::optional, IDS_HELP_ECHO_ARG1, NULL))
		);

	commandList.Add(
		L"errmsg",
		std::shared_ptr<Command>((new Command_ErrMsg())
			->AddParam(0, Command::mandatory, IDS_HELP_ERRMSG_ARG1, noListValidator)) // err code
		);

	commandList.Add(
		L"show tables",
		std::shared_ptr<Command>(((new Command_ShowTables())))
		);

	commandList.Add(
		L"show columns",
		std::shared_ptr<Command>(((
		new Command_ShowColumns())
			->AddKeyword(L"for", Command::optional)
			->AddParam(0, Command::mandatory, IDS_HELP_SHOWCOLUMNS_ARG1, tableNameValidator)
		))
		);

	commandList.Add(
		L"license details",
		std::shared_ptr<Command>(((
			new Command_LicenseDetails())
			))
		);

	commandList.Add(
		L"license add",
		std::shared_ptr<Command>(((
			new Command_LicenseAdd())
				->AddParam(0, Command::mandatory, IDS_HELP_LICENSEADD_ARG1, NULL)
			))
		);

	commandList.Add(
		L"license replace",
		std::shared_ptr<Command>(((
			new Command_LicenseAdd(true))
				->AddParam(0, Command::mandatory, IDS_HELP_LICENSEREPLACE_ARG1, NULL)
			))
		);

	commandList.Add(
		L"license",
		std::shared_ptr<Command>(((
			new Command_License())
			))
		);

	commandList.Add(
		L"show",
		std::shared_ptr<Command>((
			new Command_Show(commandFlagsStack))
				->AddParam(0, Command::mandatory, IDS_HELP_SHOW_ARG1, tableNameValidator) // table name
				->AddParam(1, Command::mandatory, IDS_HELP_SHOW_ARG2, columnForTable0Validator_selectableOnly) // columns
				->AddKeyword(L"where", Command::optional)
					->AddParam(2, Command::conditional, IDS_HELP_SHOW_ARG_WHERE, columnsWithValuesValidator)
				->AddKeyword(L"group", Command::optional)
					->AddKeyword(L"by", Command::conditional)
					->AddParam(3, Command::conditional, IDS_HELP_SHOW_ARG_GROUP_BY, columnForTable0Validator_selectableOnly)
				->AddKeyword(L"order", Command::optional)
					->AddKeyword(L"by", Command::conditional)
					->AddParam(4, Command::conditional, IDS_HELP_SHOW_ARG_ORDER_BY, columnForTable0Validator_selectableOnly)
				->AddKeyword(L"skip", Command::optional)
					->AddParam(5, Command::conditional, IDS_HELP_SHOW_ARG_SKIP, noListValidator)
				->AddKeyword(L"limit", Command::optional)
					->AddParam(6, Command::conditional, IDS_HELP_SHOW_ARG_LIMIT, noListValidator)
		));

	{
		std::shared_ptr<Command> command((
			new Command_TaskConfig())
				->AddParam(0, Command::mandatory, IDS_HELP_TASKCONFIG_CONFIGFILE, noListValidator)
		);

		AddParams_taskParams(*command, 1, noListValidator);

		commandList.Add(
			L"task config",
			command
		);
	}

	{
		std::shared_ptr<Command> command(
			new Command_TaskUpdate()
		);

		AddParams_taskParams(*command, 0, noListValidator);

		command
			->AddKeyword(L"exclude", Command::optional)
				->AddParam(5, Command::conditional, IDS_HELP_TASKUPDATE_EXCLUDE, NULL)
			->AddKeyword(L"windows_profile", Command::optional)
				->AddParam(6, Command::conditional, IDS_HELP_TASKUPDATE_WINDOWS_PROFILE, noListValidator)
			->AddKeyword(L"max_delay", Command::optional)
				->AddParam(7, Command::conditional, IDS_HELP_TASKUPDATE_MAX_DELAY, noListValidator)
		;

		commandList.Add(
			L"task update",
			command
		);
	}

	{
		std::shared_ptr<Command> command(
			new Command_TaskScan()
		);

		AddParams_taskParams(*command, 0, noListValidator);

		command
			->AddKeyword(L"exclude", Command::optional)
				->AddParam(5, Command::conditional, IDS_HELP_TASKSCAN_EXCLUDE, NULL)
			->AddKeyword(L"windows_profile", Command::optional)
				->AddParam(6, Command::conditional, IDS_HELP_TASKSCAN_PROFILE, std::make_shared<ParamValidatorEnum>(L"!InDepthScan,!MyProfile,!SmartScan,!ContextMenuScan", false, false))
			->AddKeyword(L"windows_targets", Command::optional)
				->AddParam(7, Command::conditional, IDS_HELP_TASKSCAN_WINDOWS_TARGETS, std::make_shared<ParamValidatorEnum>(L"!Memory,!RemovableDrivesBoot,!RemovableDrives,!LocalDrivesBoot,!LocalDrives,!RemoteDrives,!AllDrivesBoot,!AllDrives", true, false))
			->AddKeyword(L"windows_no_cleaning", Command::optional)
				->AddParam(8, Command::conditional, IDS_HELP_TASKSCAN_NO_CLEANING, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			->AddKeyword(L"windows_shutdown_after_scan", Command::optional)
				->AddParam(9, Command::conditional, IDS_HELP_TASKSCAN_SHUTDOWN_AFTER_SCAN, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			->AddKeyword(L"windows_allow_shutdown_cancel", Command::optional)
				->AddParam(10, Command::conditional, IDS_HELP_TASKSCAN_ALLOW_SHUTDOWN_CANCEL, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			->AddKeyword(L"linux3_targets", Command::optional)
				->AddParam(11, Command::conditional, IDS_HELP_TASKSCAN_LINUX3_TARGETS, NULL)
			->AddKeyword(L"linux3_no_cleaning", Command::optional)
				->AddParam(12, Command::conditional, IDS_HELP_TASKSCAN_NO_CLEANING, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			->AddKeyword(L"linux_profile", Command::optional)
				->AddParam(13, Command::conditional, IDS_HELP_TASKSCAN_PROFILE, std::make_shared<ParamValidatorEnum>(L"!InDepthScan,!MyProfile,!SmartScan,!ContextMenuScan", false, false))
			->AddKeyword(L"linux_targets", Command::optional)
				->AddParam(14, Command::conditional, IDS_HELP_TASKSCAN_LINUX_TARGETS, NULL)
			->AddKeyword(L"linux_no_cleaning", Command::optional)
				->AddParam(15, Command::conditional, IDS_HELP_TASKSCAN_NO_CLEANING, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			->AddKeyword(L"mobile_targets", Command::optional)
				->AddParam(16, Command::conditional, IDS_HELP_TASKSCAN_MOBILE_TARGETS, std::make_shared<ParamValidatorEnum>(L"!All", true, false))
			->AddKeyword(L"mobile_no_cleaning", Command::optional)
				->AddParam(17, Command::conditional, IDS_HELP_TASKSCAN_NO_CLEANING, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			->AddKeyword(L"max_delay", Command::optional)
				->AddParam(18, Command::conditional, IDS_HELP_TASKSCAN_MAX_DELAY, noListValidator)
		;

		commandList.Add(
			L"task scan",
			command
		);
	}

	commandList.Add(
		L"report generate",
		std::shared_ptr<Command>(((
			new Command_ReportGenerate())
				->AddParam(0, Command::mandatory, IDS_HELP_REPORT_GENERATE_ID, noListValidator))
				->AddParam(1, Command::mandatory, IDS_HELP_REPORT_GENERATE_DIRECTORY, noListValidator)
				->AddKeyword(L"type", Command::optional)
					->AddParam(2, Command::conditional, IDS_HELP_REPORT_GENERATE_TYPE, std::make_shared<ParamValidatorEnum>(L"html,zip,pdf", false)
			))
		);

	commandList.Add(
		L"report server",
		std::shared_ptr<Command>(((
			new Command_ReportServer())
				->AddParam(0, Command::mandatory, IDS_HELP_REPORT_SERVER_ID, noListValidator)
				->AddParam(1, Command::mandatory, IDS_HELP_REPORT_SERVER_DIRECTORY, noListValidator)
			))
		);

	commandList.Add(
		L"report import",
		std::shared_ptr<Command>(((
			new Command_ReportImport())
				->AddParam(0, Command::mandatory, IDS_HELP_REPORT_IMPORT_TYPE, std::make_shared<ParamValidatorEnum>(L"static,dashboard", false))
				->AddParam(1, Command::mandatory, IDS_HELP_REPORT_IMPORT_FILE, noListValidator)
			))
		);

	commandList.Add(
		L"report export",
		std::shared_ptr<Command>(((
			new Command_ReportExport())
				->AddParam(0, Command::mandatory, IDS_HELP_REPORT_EXPORT_TYPE, std::make_shared<ParamValidatorEnum>(L"static,dashboard", false))
				->AddParam(1, Command::mandatory, IDS_HELP_REPORT_EXPORT_ID, NULL)
				->AddParam(2, Command::mandatory, IDS_HELP_REPORT_EXPORT_FILE, noListValidator)
			))
		);

	commandList.Add(
		L"report create",
		std::shared_ptr<Command>(((
			new Command_ReportCreate())
				->AddParam(0, Command::mandatory, IDS_HELP_REPORT_CREATE_TYPE, std::make_shared<ParamValidatorEnum>(L"static,dashboard", false))
				->AddParam(1, Command::mandatory, IDS_HELP_REPORT_CREATE_REPORT_NAME, noListValidator)
				->AddParam(2, Command::mandatory, IDS_HELP_REPORT_CREATE_CONFIG_XML, noListValidator)
				->AddKeyword(L"active", Command::optional)
					->AddParam(3, Command::conditional, IDS_HELP_REPORT_CREATE_ACTIVE, std::make_shared<ParamValidatorEnum>(L"true,false", false))
				->AddKeyword(L"description", Command::optional)
					->AddParam(4, Command::conditional, IDS_HELP_REPORT_CREATE_DESC, noListValidator)
			))
		);

	commandList.Add(
		L"report update",
		std::shared_ptr<Command>(((
			new Command_ReportUpdate())
				->AddParam(0, Command::mandatory, IDS_HELP_REPORT_UPDATE_REPORT_ID, noListValidator)
				->AddKeyword(L"configXML", Command::optional)
					->AddParam(1, Command::conditional, IDS_HELP_REPORT_UPDATE_CONFIG_XML, noListValidator)
				->AddKeyword(L"active", Command::optional)
					->AddParam(2, Command::conditional, IDS_HELP_REPORT_UPDATE_ACTIVE, std::make_shared<ParamValidatorEnum>(L"true,false", false))
				->AddKeyword(L"description", Command::optional)
					->AddParam(3, Command::conditional, IDS_HELP_REPORT_UPDATE_DESC, noListValidator)
			))
		);

	commandList.Add(
		L"report delete",
		std::shared_ptr<Command>(((
			new Command_ReportDelete())
				->AddParam(0, Command::mandatory, IDS_HELP_REPORT_DELETE_REPORT_ID, noListValidator)
			))
		);

	commandList.Add(
		L"report",
		std::shared_ptr<Command>(((
			new Command_Report(commandFlagsStack))
				->AddParam(0, Command::mandatory, IDS_HELP_REPORT_TYPE, std::make_shared<ParamValidatorEnum>(L"static,dashboard,generated", false))
			))
		);

	commandList.Add(
		L"request",
		std::shared_ptr<Command>((
			new Command_Request())
				->AddParam(0, Command::mandatory, IDS_HELP_REQUEST_DATATYPELIST, std::make_shared<ParamValidatorEnum>(L"sysinspector,configuration,protection_status,protection_features,system_information", true))
				->AddParam(1, Command::mandatory, IDS_HELP_REQUEST_CLIENTS, NULL)
				->AddKeyword(L"si_compare", Command::optional)
					->AddParam(2, Command::conditional, IDS_HELP_REQUEST_SICOMPARE, noListValidator)
				->AddParam(3, Command::optional, IDS_HELP_REQUEST_SISNAPSHOT, std::make_shared<ParamValidatorEnum>(L"si_snapshot", false))
		));


	commandList.Add(
		L"rule create",
		std::shared_ptr<Command>(
			(new Command_PolicyRuleCreate())
			->AddParam(0, Command::mandatory, IDS_HELP_POLICY_RULE_CREATE_XML, noListValidator)
			->AddParam(1, Command::mandatory, IDS_HELP_POLICY_RULE_CREATE_NAME, noListValidator)			
			->AddParam(2, Command::mandatory, IDS_HELP_POLICY_RULE_CREATE_POLICYID, std::make_shared<ParamValidatorEnum>(L"!DefaultClientsPolicy", false, false))
			->AddKeyword(L"desc", Command::optional)
				->AddParam(3, Command::conditional, IDS_HELP_POLICY_RULE_CREATE_DESC, noListValidator)
			->AddKeyword(L"priority", Command::optional)
				->AddParam(4, Command::conditional, IDS_HELP_POLICY_RULE_CREATE_PRIORITY, std::make_shared<ParamValidatorEnum>(L"top,bottom", false))
			->AddKeyword(L"enabled", Command::optional)
				->AddParam(5, Command::conditional, IDS_HELP_POLICY_RULE_CREATE_ENABLED, std::make_shared<ParamValidatorEnum>(L"true,false", false))
		));

	commandList.Add(
		L"rule delete",
		std::shared_ptr<Command>(((
			new Command_PolicyRuleDelete())
				->AddParam(0, Command::mandatory, IDS_HELP_POLICY_RULE_DELETE_RULE_ID, noListValidator)
			))
		);

	commandList.Add(
		L"rule import",
		std::shared_ptr<Command>(((
			new Command_PolicyRuleImport())
				->AddParam(0, Command::mandatory, IDS_HELP_POLICY_RULE_IMPORT_FILE, noListValidator)
			))
		);

	commandList.Add(
		L"rule export",
		std::shared_ptr<Command>(((
			new Command_PolicyRuleExport())
				->AddParam(0, Command::mandatory, IDS_HELP_POLICY_RULE_EXPORT_ID, NULL)
				->AddParam(1, Command::mandatory, IDS_HELP_POLICY_RULE_EXPORT_FILE, noListValidator)
			))
		);

	commandList.Add(
		L"rule update",
		std::shared_ptr<Command>(((
			new Command_PolicyRuleUpdate())
				->AddParam(0, Command::mandatory, IDS_HELP_POLICY_RULE_UPDATE_RULE_ID, noListValidator)
				->AddKeyword(L"xml", Command::optional)
					->AddParam(1, Command::conditional, IDS_HELP_POLICY_RULE_UPDATE_XML, noListValidator)
				->AddKeyword(L"desc", Command::optional)
					->AddParam(2, Command::conditional, IDS_HELP_POLICY_RULE_UPDATE_DESC, noListValidator)
				->AddKeyword(L"policy", Command::optional)
					->AddParam(3, Command::conditional, IDS_HELP_POLICY_RULE_UPDATE_POLICYID, std::make_shared<ParamValidatorEnum>(L"!DefaultClientsPolicy", false, false))
				->AddKeyword(L"priority", Command::optional)
					->AddParam(4, Command::conditional, IDS_HELP_POLICY_RULE_UPDATE_PRIORITY, std::make_shared<ParamValidatorEnum>(L"up,down,top,bottom", false))
				->AddKeyword(L"enabled", Command::optional)
					->AddParam(5, Command::conditional, IDS_HELP_POLICY_RULE_UPDATE_ENABLED, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			))
		);

	commandList.Add(
		L"rule",
		std::shared_ptr<Command>((new Command_PolicyRules()))
		);

	commandList.Add(
		L"policy import",
		std::shared_ptr<Command>(((
		new Command_PolicyImport())
		->AddParam(0, Command::mandatory, IDS_HELP_POLICY_IMPORT_FILE, noListValidator)		
		))
		);

	commandList.Add(
		L"policy export",
		std::shared_ptr<Command>(((
		new Command_PolicyExport())
		->AddParam(0, Command::mandatory, IDS_HELP_POLICY_EXPORT_ID, std::make_shared<ParamValidatorEnum>(L"!DefaultClientsPolicy,*", false, false))
		->AddParam(1, Command::mandatory, IDS_HELP_POLICY_EXPORT_FILE, noListValidator)
		->AddParam(2, Command::optional, IDS_HELP_POLICY_EXPORT_TREE, std::make_shared<ParamValidatorEnum>(L"true,false", false))
		))
		);

	commandList.Add(
		L"policy assign",
		std::shared_ptr<Command>(
		(new Command_PolicyAssign())
		->AddParam(0, Command::mandatory, IDS_HELP_POLICY_ASSIGN_POLICY_ID, std::make_shared<ParamValidatorEnum>(L"!DefaultClientsPolicy", false, false))
		->AddParam(1, Command::mandatory, IDS_HELP_POLICY_ASSIGN_CLIENT_ID_LIST, NULL)
		)
	);

	commandList.Add(
		L"policy create",
		std::shared_ptr<Command>((new Command_PolicyCreate())
				->AddParam(0, Command::mandatory, IDS_HELP_POLICY_CREATE_POLICY_NAME, noListValidator)
				->AddParam(1, Command::mandatory, IDS_HELP_POLICY_CREATE_CONFIG_XML, noListValidator)
			->AddKeyword(L"parentID", Command::optional)
				->AddParam(2, Command::conditional, IDS_HELP_POLICY_CREATE_POLICY_PARENT_ID, noListValidator)
			->AddKeyword(L"description", Command::optional)
				->AddParam(3, Command::conditional, IDS_HELP_POLICY_CREATE_POLICY_DESCRIPTION, noListValidator)
			->AddKeyword(L"overrideAnyChild", Command::optional)
				->AddParam(4, Command::conditional, IDS_HELP_POLICY_CREATE_OVERRIDE_ANY_CHILD, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			->AddKeyword(L"downReplicable", Command::optional)
				->AddParam(5, Command::conditional, IDS_HELP_POLICY_CREATE_DOWN_REPLICABLE, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			->AddKeyword(L"defaultForClients", Command::optional)
				->AddParam(6, Command::conditional, IDS_HELP_POLICY_CREATE_DEFAULT_FOR_CLIENTS, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			->AddKeyword(L"defaultForLowerServers", Command::optional)
				->AddParam(7, Command::conditional, IDS_HELP_POLICY_CREATE_DEFAULT_FOR_LOWER_SERVERS, std::make_shared<ParamValidatorEnum>(L"true,false", false))
	));

	commandList.Add(
		L"policy update",
		std::shared_ptr<Command>((new Command_PolicyUpdate())
				->AddParam(0, Command::mandatory, IDS_HELP_POLICY_UPDATE_POLICY_ID, noListValidator)
			->AddKeyword(L"name", Command::optional)
				->AddParam(1, Command::conditional, IDS_HELP_POLICY_UPDATE_POLICY_NAME, noListValidator)
			->AddKeyword(L"parentID", Command::optional)
				->AddParam(2, Command::conditional, IDS_HELP_POLICY_UPDATE_POLICY_PARENT_ID, noListValidator)
			->AddKeyword(L"configXML", Command::optional)
				->AddParam(3, Command::conditional, IDS_HELP_POLICY_UPDATE_CONFIG_XML, noListValidator)
			->AddKeyword(L"description", Command::optional)
				->AddParam(4, Command::conditional, IDS_HELP_POLICY_UPDATE_POLICY_DESCRIPTION, noListValidator)
			->AddKeyword(L"overrideAnyChild", Command::optional)
				->AddParam(5, Command::conditional, IDS_HELP_POLICY_UPDATE_OVERRIDE_ANY_CHILD, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			->AddKeyword(L"downReplicable", Command::optional)
				->AddParam(6, Command::conditional, IDS_HELP_POLICY_UPDATE_DOWN_REPLICABLE, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			->AddKeyword(L"defaultForClients", Command::optional)
				->AddParam(7, Command::conditional, IDS_HELP_POLICY_UPDATE_DEFAULT_FOR_CLIENTS, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			->AddKeyword(L"defaultForLowerServers", Command::optional)
				->AddParam(8, Command::conditional, IDS_HELP_POLICY_UPDATE_DEFAULT_FOR_LOWER_SERVERS, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			->AddKeyword(L"replicated_clients", Command::optional)
				->AddParam(9, Command::conditional, IDS_HELP_POLICY_UPDATE_REPLICATED_CLIENTS, noListValidator)
			->AddKeyword(L"lower_servers_default", Command::optional)
				->AddParam(10, Command::conditional, IDS_HELP_POLICY_UPDATE_LOWER_SERVERS_DEFAULT, std::make_shared<ParamValidatorEnum>(L"!NotAvailable", false, false))
			->AddKeyword(L"primary_clients_default", Command::optional)
				->AddParam(11, Command::conditional, IDS_HELP_POLICY_UPDATE_PRIMARY_CLIENTS_DEFAULT, noListValidator)
	));

	commandList.Add(
		L"policy delete",
		std::shared_ptr<Command>((new Command_PolicyDelete())
				->AddParam(0, Command::mandatory, IDS_HELP_POLICY_DELETE_POLICY_ID, noListValidator)
			->AddKeyword(L"child_policies", Command::optional)
				->AddParam(1, Command::conditional, IDS_HELP_POLICY_DELETE_CHILD_POLICIES, std::make_shared<ParamValidatorEnum>(L"!DefaultUpperServerPolicy,!NotAvailable", false, false))
			->AddKeyword(L"primary_clients", Command::optional)
				->AddParam(2, Command::conditional, IDS_HELP_POLICY_DELETE_PRIMARY_CLIENTS, std::make_shared<ParamValidatorEnum>(L"!DefaultClientsPolicy", false, false))
			->AddKeyword(L"replicated_clients", Command::optional)
				->AddParam(3, Command::conditional, IDS_HELP_POLICY_DELETE_REPLICATED_CLIENTS, noListValidator)
			->AddKeyword(L"primary_clients_default", Command::optional)
				->AddParam(4, Command::conditional, IDS_HELP_POLICY_DELETE_PRIMARY_CLIENTS_DEFAULT, noListValidator)
			->AddKeyword(L"lower_servers_default", Command::optional)
				->AddParam(5, Command::conditional, IDS_HELP_POLICY_DELETE_LOWER_SERVERS_DEFAULT, std::make_shared<ParamValidatorEnum>(L"!NotAvailable", false, false))
			->AddKeyword(L"whole_branch", Command::optional)
				->AddParam(6, Command::conditional, IDS_HELP_POLICY_DELETE_WHOLE_BRANCH, std::make_shared<ParamValidatorEnum>(L"true,false", false))
	));

	commandList.Add(
		L"policy",
		std::shared_ptr<Command>((new Command_Policy())
			->AddParam(0, Command::optional, IDS_HELP_POLICY_TREE, std::make_shared<ParamValidatorEnum>(L"tree", false))
	));

	commandList.Add(
		L"servercfg get",
		std::shared_ptr<Command>(((
			new Command_ServerConfig(GET_SERVER_CFG))
				->AddParam(0, Command::mandatory, IDS_HELP_SERVERCFG_GET_ARG1, noListValidator)
			))
		);

	commandList.Add(
		L"servercfg put",
		std::shared_ptr<Command>(((
			new Command_ServerConfig(PUT_SERVER_CFG))
				->AddParam(0, Command::mandatory, IDS_HELP_SERVERCFG_PUT_ARG1, noListValidator)
			))
		);

	commandList.Add(
		L"servercfg set",
		std::shared_ptr<Command>(((
			new Command_ServerConfig(SET_SERVER_CFG))
				->AddParam(0, Command::mandatory, IDS_HELP_SERVERCFG_SET_ARG1, serverCfgItemsValidator)
			))
		);

	commandList.Add(
		L"servercfg setpwd",
		std::shared_ptr<Command>(((
			new Command_ServerConfig(SETPWD_SERVER_CFG, this))
				->AddParam(0, Command::mandatory, IDS_HELP_SERVERCFG_SETPWD_ARG1, serverCfgPwdItemsValidator)
			))
		);

	commandList.Add(
		L"servercfg list",
		std::shared_ptr<Command>(((
			new Command_ServerConfig(LIST_SERVER_CFG))
			))
		);

	commandList.Add(
		L"history",
		std::shared_ptr<Command>(((
			new Command_History())
			->AddParam(0, Command::optional, IDS_HELP_HISTORY_ARG1, std::make_shared<ParamValidatorEnum>(L"true,false,clear,list", false))
			))
		);

	commandList.Add(
		L"path",
		std::shared_ptr<Command>(((
			new Command_Path(scriptPath))
				->AddParam(0, Command::optional, IDS_HELP_PATH_ARG1, std::make_shared<ParamValidatorEnum>(L"get,set,script", false))
				->AddParam(1, Command::optional, IDS_HELP_PATH_ARG2, noListValidator)
			))
		);

	commandList.Add(
		L"client rename",
		std::shared_ptr<Command>(((
			new Command_ClientRename())
				->AddParam(0, Command::mandatory, IDS_HELP_CLIENT_RENAME_CLIENT_ID, noListValidator)
				->AddParam(1, Command::mandatory, IDS_HELP_CLIENT_RENAME_CLIENT_NAME, noListValidator)
			))
	);

	commandList.Add(
		L"client comment",
		std::shared_ptr<Command>(((
			new Command_ClientComment())
				->AddParam(0, Command::mandatory, IDS_HELP_CLIENT_COMMENT_CLIENT_ID, noListValidator)
				->AddParam(1, Command::mandatory, IDS_HELP_CLIENT_COMMENT_CLIENT_COMMENT, noListValidator)
			))
	);

	commandList.Add(
		L"password",
		std::shared_ptr<Command>( ((
		new Command_Password(this))
		->AddParam(0, Command::mandatory, IDS_HELP_PASSWORD_TYPE, std::make_shared<ParamValidatorEnum>(L"client,replication,installer,currentuser", false))
		->AddParam(1, Command::optional, IDS_HELP_PASSWORD_OLD, NULL)
		->AddParam(2, Command::conditional, IDS_HELP_PASSWORD_NEW, NULL)
		))
		);

	commandList.Add(
		L"encrypt",
		std::shared_ptr<Command>( ((
		new Command_Encrypt(this))
		->AddParam(0, Command::mandatory, IDS_HELP_ENCRYPTION_TYPE, std::make_shared<ParamValidatorEnum>(L"server,other,hash,int64", false))
		->AddParam(1, Command::optional, IDS_HELP_PASSWORD_TO_ENCRYPT, NULL)
		))
		);

	commandList.Add(
		L"getdata",
		std::shared_ptr<Command>( ((
			new Command_GetData())
				->AddParam(0, Command::mandatory, IDS_HELP_GETDATA_TYPE, std::make_shared<ParamValidatorEnum>(L"sysinspector,configuration,protection_status,protection_features,system_information,policy,policy_merged,policy_override,policy_nonoverride", true))
				->AddParam(1, Command::mandatory, IDS_HELP_GETDATA_DATA_ID, noListValidator)
				->AddParam(2, Command::mandatory, IDS_HELP_GETDATA_FILE, noListValidator)
		))
		);

	commandList.Add(
		L"group import",
		std::shared_ptr<Command>(((
			new Command_GroupImport())
				->AddParam(0, Command::mandatory, IDS_HELP_GROUP_IMPORT_TYPE, std::make_shared<ParamValidatorEnum>(L"static,parametric", false))
				->AddParam(1, Command::mandatory, IDS_HELP_GROUP_IMPORT_ID, noListValidator)
				->AddParam(2, Command::mandatory, IDS_HELP_GROUP_IMPORT_FILE, noListValidator)
				->AddParam(3, Command::optional, IDS_HELP_GROUP_IMPORT_CLIENT_RELATIONS, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			))
		);

	commandList.Add(
		L"group export",
		std::shared_ptr<Command>(((
			new Command_GroupExport())
				->AddParam(0, Command::mandatory, IDS_HELP_GROUP_EXPORT_TYPE, std::make_shared<ParamValidatorEnum>(L"static,parametric", false))
				->AddParam(1, Command::mandatory, IDS_HELP_GROUP_EXPORT_ID, NULL)
				->AddParam(2, Command::mandatory, IDS_HELP_GROUP_EXPORT_FILE, noListValidator)
				->AddParam(3, Command::optional, IDS_HELP_GROUP_EXPORT_TREE, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			))
		);

	commandList.Add(
		L"group create",
		std::shared_ptr<Command>(((
			new Command_GroupCreate())
				->AddParam(0, Command::mandatory, IDS_HELP_GROUP_CREATE_GROUP_TYPE, std::make_shared<ParamValidatorEnum>(L"static,parametric", false))
				->AddParam(1, Command::mandatory, IDS_HELP_GROUP_CREATE_GROUP_NAME, noListValidator)
				->AddKeyword(L"description", Command::optional)
					->AddParam(2, Command::conditional, IDS_HELP_GROUP_CREATE_GROUP_DESCRIPTION, noListValidator)
				->AddKeyword(L"parentID", Command::optional)
					->AddParam(3, Command::conditional, IDS_HELP_GROUP_CREATE_GROUP_PARENT_ID, noListValidator)
				->AddKeyword(L"paramsXML", Command::optional)
					->AddParam(4, Command::conditional, IDS_HELP_GROUP_CREATE_PARAMS_XML, noListValidator)
				->AddKeyword(L"sticky", Command::optional)
					->AddParam(5, Command::conditional, IDS_HELP_GROUP_CREATE_STICKY, std::make_shared<ParamValidatorEnum>(L"true,false", false))
			))
		);

	commandList.Add(
		L"group delete",
		std::shared_ptr<Command>(((
			new Command_GroupDelete())
				->AddParam(0, Command::mandatory, IDS_HELP_GROUP_DELETE_GROUP_ID, noListValidator)
			))
		);

	commandList.Add(
		L"group assign",
		std::shared_ptr<Command>((
			new Command_GroupAssign())
				->AddParam(0, Command::mandatory, IDS_HELP_GROUP_ASSIGN_GROUP_ID, noListValidator)
				->AddParam(1, Command::mandatory, IDS_HELP_GROUP_ASSIGN_CLIENT_ID_LIST, NULL)
		)
	);

	commandList.Add(
		L"group remove",
		std::shared_ptr<Command>((
			new Command_GroupRemove())
				->AddParam(0, Command::mandatory, IDS_HELP_GROUP_REMOVE_GROUP_ID, noListValidator)
				->AddParam(1, Command::mandatory, IDS_HELP_GROUP_REMOVE_CLIENT_ID_LIST, NULL)
			)
	);

	commandList.Add(
		L"group update",
		std::shared_ptr<Command>(((
			new Command_GroupUpdate())
				->AddParam(0, Command::mandatory, IDS_HELP_GROUP_UPDATE_GROUP_ID, noListValidator)
				->AddKeyword(L"name", Command::optional)
					->AddParam(1, Command::conditional, IDS_HELP_GROUP_UPDATE_GROUP_NAME, noListValidator)
				->AddKeyword(L"description", Command::optional)
					->AddParam(2, Command::conditional, IDS_HELP_GROUP_UPDATE_GROUP_DESCRIPTION, noListValidator)
				->AddKeyword(L"parentID", Command::optional)
					->AddParam(3, Command::conditional, IDS_HELP_GROUP_UPDATE_GROUP_PARENT_ID, noListValidator)
				->AddKeyword(L"paramsXML", Command::optional)
					->AddParam(4, Command::conditional, IDS_HELP_GROUP_UPDATE_PARAMS_XML, noListValidator)
			))
		);

	commandList.Add(
		L"group",
		std::shared_ptr<Command>((new Command_Group(commandFlagsStack))
			->AddParam(0, Command::optional, IDS_HELP_GROUP_TREE, std::make_shared<ParamValidatorEnum>(L"tree", false))
	));

	commandList.Add(
		L"scanlog",
		std::shared_ptr<Command>(((
			new Command_ScanLog(commandFlagsStack))
				->AddParam(0, Command::mandatory, IDS_HELP_SCANLOG_ARG1,noListValidator)
			))
		);

	commandList.Add(
		L"logforward",
		std::shared_ptr<Command>(((
			new Command_LogForward(commandFlagsStack))
				->AddParam(0, Command::optional, IDS_HELP_LOGFORWARD_ARG1, std::make_shared<ParamValidatorEnum>(L"event,threat,firewall,hips,antispam,greylist,scan,mobile,device_control,web_control", false))
				->AddParam(1, Command::optional, IDS_HELP_LOGFORWARD_ARG2, std::make_shared<ParamValidatorEnumConditional>(0, L"true,false", false))
				->AddKeyword(L"level", Command::optional, 2)
					->AddParam(2, Command::conditional, IDS_HELP_LOGFORWARD_ARG3, std::make_shared<ParamValidatorEnumConditional>(1, L"critical,warning,normal,diagnostics", false))
				->AddKeyword(L"severity", Command::optional, 2)
					->AddParam(3, Command::conditional, IDS_HELP_LOGFORWARD_ARG4, std::make_shared<ParamValidatorEnumConditional>(1, L"informational,error,warning,debug", false))
				->AddKeyword(L"facility", Command::optional, 2)
					->AddParam(4, Command::conditional, IDS_HELP_LOGFORWARD_ARG5, noListValidator)
			))
		);

	commandList.Add(
		L"client new",
		std::shared_ptr<Command>(((
			new Command_ClientNew())
				->AddParam(0, Command::mandatory, IDS_HELP_CLIENT_NEW_CLIENT_ID, NULL)
				->AddParam(1, Command::mandatory, IDS_HELP_CLIENT_NEW_FLAG, std::make_shared<ParamValidatorEnum>(L"set,reset", false))
			))
		);

	commandList.Add(
		L"client delete",
		std::shared_ptr<Command>(((
			new Command_ClientDelete())
				->AddParam(0, Command::mandatory, IDS_HELP_CLIENT_DELETE_CLIENT_ID, NULL)
				->AddParam(1, Command::optional, IDS_HELP_CLIENT_DELETE_NOWAIT, std::make_shared<ParamValidatorEnum>(L"nowait", false))
			))
		);

	commandList.Add(
		L"client roaming",
		std::shared_ptr<Command>(((
			new Command_ClientRoaming())
				->AddParam(0, Command::mandatory, IDS_HELP_CLIENT_ROAMING_CLIENT_ID, NULL)
				->AddParam(1, Command::mandatory, IDS_HELP_CLIENT_ROAMING_FLAG, std::make_shared<ParamValidatorEnum>(L"set,reset", false))
			))
		);

	commandList.Add(
		L"restart",
		std::shared_ptr<Command>((new Command_Restart(*this))
			->AddParam(0, Command::optional, IDS_HELP_RESTART_ARG1, std::make_shared<ParamValidatorEnum>(L"full", false)))
		);

	commandList.Add(
		L"clearinfo",
		std::shared_ptr<Command>((
			new Command_ClearInfo())
				->AddParam(0, Command::mandatory, IDS_HELP_CLEARINFO_DATATYPELIST, std::make_shared<ParamValidatorEnum>(L"threat,firewall,event,scan,custom", true))
				->AddParam(1, Command::mandatory, IDS_HELP_CLEARINFO_CLIENTS, NULL)
		));

	commandList.Add(
		L"runnow",
		std::shared_ptr<Command>((
			new Command_RunNow())
				->AddParam(0, Command::mandatory, IDS_HELP_RUNNOW_ACTIONS, std::make_shared<ParamValidatorEnum>(L"cleanup,compact,replicate,replicate_with_mark_all_clients,update,update_with_clear_cache,apply_policy_rules,synchronize_parametric_groups", true))
		));

//---------------------------------------------------------------------------------------------------------------
// THIS HAS TO BE LAST COMMAND!!!
	commandList.Add(
		L"help",
		std::shared_ptr<Command>((new Command_Help(commandList))
		->AddParam(0, Command::optional, IDS_HELP_HELP_ARG1, std::make_shared<ParamValidatorHelp>(commandList))
		->AddParam(1, Command::optional, IDS_HELP_HELP_ARG2, noListValidator))
		);
}

ScriptInterpreter::~ScriptInterpreter()
{
	commandFlagsStack.Pop();
	if (connectedByThis)
	{
		Disconnect();
		connectedByThis = false;
	}
}

void ScriptInterpreter::AddParams_taskParams(Command &command, const int offset, std::shared_ptr<ParamValidator> noListValidator)
{
	(&command)
		->AddParam(offset + 0, Command::mandatory, IDS_HELP_TASK_CLIENTS, NULL)
		->AddKeyword(L"name", Command::optional)
			->AddParam(offset + 1, Command::conditional, IDS_HELP_TASK_NAME, noListValidator)
		->AddKeyword(L"description", Command::optional)
			->AddParam(offset + 2, Command::conditional, IDS_HELP_TASK_DESCRIPTION, noListValidator)
		->AddKeyword(L"applyAfter", Command::optional)
			->AddParam(offset + 3, Command::conditional, IDS_HELP_TASK_APPLY_AFTER, noListValidator)
		->AddKeyword(L"deleteIfCompleted", Command::optional)
			->AddParam(offset + 4, Command::conditional, IDS_HELP_TASK_DELETE_IF_COMPLETED, std::make_shared<ParamValidatorEnum>(L"true,false", false))
	;
}

CommandList *ScriptInterpreter::GetCommandList()
{
	return &commandList;
}

bool ScriptInterpreter::ExecuteScript(const std::wstring &scriptSrc, bool connect, AutocompleteCollector *collector)
{
	CommandListParser commandListParser(scriptSrc);
	StringVector commands = commandListParser.GetCommands();

	if ((commands.size() == 0) && connect)
	{
		// connected in the handling of the first command (so must have at least one command)
		commands.push_back(L"");
	}

	bool commandExecuted = false;

	for (StringVector::const_iterator it = commands.begin(); it != commands.end(); ++it)
	{
		ScriptInterpreter::lastCommand = *it;

		std::wstring replacedCommand;
		ReplaceAtFileContent(*it, replacedCommand);

		ScriptInterpreter::lastCommand = replacedCommand;

		ParseCommand(replacedCommand);

		if (connect)
		{
			try
			{
				// connect to server, take parameters from tokens array and remove them from there
				Connect();
				connectedByThis = true;
				connect = false;
			}
			catch (EraCMDException &e)
			{
				e.command.clear();
				throw;
			}
		}

		if (tokens.empty())
			return false;

		if (ExecuteCommand(tokens.begin(), tokens.end(), collector))
			commandExecuted = true;

		tokens.clear();

		if (ShouldExit())
			break;
	}

	return commandExecuted;
}


// replace @file@ with content of file
void ScriptInterpreter::ReplaceAtFileContent(const std::wstring &scriptSrc, std::wstring &outResult)
{
	FilesReplace fr(scriptSrc, 32);
	outResult = fr.ReplaceAndGetResult();
}


// convert input to tokens
void ScriptInterpreter::ParseCommand(const std::wstring &scriptSrc)
{
	tokens.clear();
	CommandParser parser;
	parser.ProcessCommandLine(scriptSrc.c_str(), tokens);

#ifdef _DEBUG
	// DumpTokenizedCommand(tokens);
#endif
}

void ScriptInterpreter::ProcessConnectionParams(ConnectionParams &connectionParams)
{
	if (connectionParams.askAll)
	{
		std::wcout << LoadStringFromResource(IDS_ENTER_ALL) << std::endl;
		ConsoleKeyboardInput keyInput(*this, false);

		// host
		std::wcout << LoadStringFromResource(IDS_ENTER_SERVER) << L" [" << connectionParams.host << L':' << connectionParams.port << L"]: ";
		keyInput.GetInput();
		std::wstring enteredHost = keyInput.GetLine();
		if (!enteredHost.empty())
		{
			int ret = ParseHost(enteredHost, connectionParams.host, connectionParams.port);
			if (ret != 0)
			{
				std::wcout << std::endl;
				throw EraCMDException_Pause(CMDERROR_INVALID_PORT, enteredHost.c_str());
			}
		}
		std::wcout << std::endl;

		// era or domain user?
		std::wcout << LoadStringFromResource(IDS_ENTER_USER_ERA_OR_DOMAIN);
		std::wstring enteredAnswer,
					 eraStr = L"era",
					 windowsStr = L"windows";
		bool answerIsEra, answerIsWindows, answerIsBlank;
		do
		{
			keyInput.GetInput();
			enteredAnswer = keyInput.GetLine();
			answerIsEra = _wcsicmp(enteredAnswer.c_str(), eraStr.c_str()) == 0;
			answerIsWindows = _wcsicmp(enteredAnswer.c_str(), windowsStr.c_str()) == 0;
			answerIsBlank = enteredAnswer == L"";
			if (answerIsEra)
				connectionParams.userDomain = false;
			else if (answerIsWindows)
				connectionParams.userDomain = true;
			else if (!answerIsBlank)
				std::wcout << std::endl << LoadStringFromResource(IDS_ENTER_NOT_UNDERSTOOD) << LoadStringFromResource(IDS_ENTER_USER_ERA_OR_DOMAIN);
		}
		while (!answerIsEra && !answerIsWindows && !answerIsBlank);
		std::wcout << std::endl;

		// user
		std::wcout << LoadStringFromResource(IDS_ENTER_USER) << L" [" << (connectionParams.userDomain ? LoadStringFromResource(IDS_WINDOWS_SESSION_USER) : connectionParams.user) << L"]: ";
		keyInput.GetInput();
		std::wstring enteredUser = keyInput.GetLine();
		if (!enteredUser.empty())
			connectionParams.user = enteredUser;
		else if (connectionParams.userDomain)
			connectionParams.userCurrent = true;
		std::wcout << std::endl;

		// password
		if (!connectionParams.userCurrent)
		{
			std::wcout << LoadStringFromResource(IDS_ENTER_PASSWORD2) << L": ";
			keyInput.SetInkognitoMode(true);
			keyInput.GetInput();
			connectionParams.password = keyInput.GetLine();
			std::wcout << std::endl;
		}
	}
	else if (connectionParams.askPassword)
	{
		ConsoleKeyboardInput keyInput(*this, true);
		std::wstring passwordMessage = LoadStringFromResource(IDS_ENTER_PASSWORD);
		std::wcout << passwordMessage << L" ";
		keyInput.GetInput();
		connectionParams.password = keyInput.GetLine();
		std::wcout << std::endl;
#ifdef _DEBUG
		// std::wcout << std::endl << L"password is set to " << connectionParams.password << std::endl;
#endif
	}
}

// connect to era server
void ScriptInterpreter::Connect()
{
	ConnectionParamsProcessor connParamsProc(tokens);
	std::pair<tokenArray::const_iterator, tokenArray::const_iterator> processedTokens;
	processedTokens = connParamsProc.ProcessTo(connectionParams);
	if (processedTokens.first != tokens.end())
		tokens.erase(processedTokens.first, processedTokens.second);

	ProcessConnectionParams(connectionParams);

#ifdef _DEBUG
	// std::wcout << connectionParams << std::endl;
#endif

	uint32_t resultFromApi = era_Connect(connectionParams.host.c_str(), connectionParams.port, connectionParams.user.c_str(), connectionParams.password.c_str(), connectionParams.userDomain, connectionParams.userCurrent);
	if (resultFromApi != 0)
	{
		if (connectionParams.askAll)
			throw EraCMDException_FromApi_Pause(resultFromApi);
		else
			throw EraCMDException_FromApi(resultFromApi);
	}
}

void ScriptInterpreter::Disconnect()
{
	(void)era_Disconnect();
}

// execute a single command
bool ScriptInterpreter::ExecuteCommand(tokenArray::iterator commandBegin, tokenArray::iterator commandEnd, AutocompleteCollector *collector)
{
	ScriptInterpreter::lastCommand = TokenArrayToString(commandBegin, commandEnd);

	const bool continueOnErrorWhileProcessingFlags = commandFlagsStack.CurrentFlags().continueOnError;

	commandFlagsStack.Push();
	scope_exit se(std::bind(&CommandFlagsStack::Pop, &commandFlagsStack));

	elementArray elements;
	bool hasFlags = false;
	try
	{
		GroupTokensIntoElements(commandBegin, commandEnd, elements, collector);

		hasFlags = ProcessAndApplyInlineFlags(elements);
		if (elements.empty()) return false; // prazdny command, ignorujem
	}
	catch (const EraCMDException &e)
	{
		if (continueOnErrorWhileProcessingFlags)
		{
			e.Print(std::wcerr);
			std::wcerr << gConsoleColors.LIGHT_RED << LoadStringFromResource(IDS_CONTINUING_EXECUTION) << gConsoleColors.DEFAULT << std::endl;
			partiallyFailed = true;
			return true;
		}
		else
		{
			throw;
		}
	}

	try
	{
		if (hasFlags && collector)
			return true; // no autocomplete after flags

	#ifdef _DEBUG
		// std::wcout << commandFlagsStack.currentFlags() << std::endl;
	#endif

		OutputHandler outputHandler(commandFlagsStack.CurrentFlags(), gConsoleColors, gTextWidth_Console, gTextWidth_File);

		if (elements[0]->type != Element::element_word)
			throw EraCMDException(CMDERROR_INVALID_COMMAND_SYNTAX, elements[0]->GetString().c_str());

		// find command in command table and execute it
		std::wstring commandString;
		std::shared_ptr<Command> command = commandList.GetCommand(elements, commandString);

		if (command == NULL)
			throw EraCMDException(CMDERROR_UNKNOWN_COMMAND, elements[0]->GetString().c_str());
		else
		{
			command->GrabArguments(elements, commandString, collector);
			if (collector == NULL)
			{
				command->Execute(elements, outputHandler);
			}
		}
	}
	catch (const EraCMDException &e)
	{
		if (commandFlagsStack.CurrentFlags().continueOnError)
		{
			e.Print(std::wcerr);
			std::wcerr << gConsoleColors.LIGHT_RED << LoadStringFromResource(IDS_CONTINUING_EXECUTION) << gConsoleColors.DEFAULT << std::endl;
			partiallyFailed = true;
		}
		else
		{
			throw;
		}
	}

	return true;
}

bool AddWordlistToElementList(std::shared_ptr<ElementWordList> inWordList, elementArray &outElements)
{
	if (inWordList->wordList.empty())
		throw std::exception("empty wordlist");

	if (inWordList->wordList.size() == 1) // only one element, add as single word
	{
		outElements.push_back(inWordList->wordList[0]);
		inWordList->wordList.clear();
		return false;
	}
	else
	{
		outElements.push_back(inWordList);
		return true;
	}
}

// group tokens to WORDS/LISTS/FLAGS
void ScriptInterpreter::GroupTokensIntoElements(tokenArray::iterator commandBegin, tokenArray::iterator commandEnd, elementArray &outElements, AutocompleteCollector *collector)
{
	tokenArray::iterator originalCommandBegin = commandBegin;

	// first create words & words with commas
	std::shared_ptr<Token> lastToken;
	std::shared_ptr<ElementWord> currentWord;
	std::shared_ptr<ElementWordList> currentWordList = std::make_shared<ElementWordList>();

	bool lastWordIsComma = false;

	while (commandBegin != commandEnd)
	{
		std::shared_ptr<Token> token = *commandBegin;
		commandBegin++;

		if (token->type == Token::token_par_left)
		{
			if ((lastToken->type != Token::token_in) && (lastToken->type != Token::token_notin))
			{
				throw EraCMDException(CMDERROR_LEFT_PARENTHESIS_ONLY_AFTER_IN);
			}

			if (commandBegin == commandEnd)
			{
				currentWord->missingOperand = true;
				if (collector)
				{
					AddWordlistToElementList(currentWordList, outElements);
					return;
				}
				throw EraCMDException(CMDERROR_MISSING_RIGHT_PARENTHESIS);
			}
			token = *commandBegin;
			commandBegin++;

			bool first = true;
			do
			{
				if (first)
				{
					first = false;
				}
				else
				{
					if (token->type != Token::token_comma)
					{
						throw std::exception("missing comma");
					}

					if (commandBegin == commandEnd)
					{
						if (collector)
						{
							currentWord->missingOperand = true;
							AddWordlistToElementList(currentWordList, outElements);
							return;
						}
						throw EraCMDException(CMDERROR_VALUE_EXPECTED);
					}
					token = *commandBegin;
					commandBegin++;
				}

				if (token->type != Token::token_text)
				{
					if (collector)
					{
						currentWord->missingOperand = true;
						AddWordlistToElementList(currentWordList, outElements);
						return;
					}
					throw EraCMDException(CMDERROR_VALUE_EXPECTED);
				}
				currentWord->tokenList.push_back(token);

				if (commandBegin == commandEnd)
				{
					currentWord->missingOperand = true;
					if (collector)
					{
						AddWordlistToElementList(currentWordList, outElements);
						return;
					}
					throw EraCMDException(CMDERROR_MISSING_RIGHT_PARENTHESIS);
				}
				token = *commandBegin;
				commandBegin++;
			} while (token->type != Token::token_par_right);

			if (commandBegin == commandEnd)
			{
				break;
			}

			lastToken = token;
			token = *commandBegin;
			commandBegin++;
		}

		// handle first token if already processed
		if (lastToken)
		{
			if (token->type == Token::token_comma)
			{
				if (currentWord->tokenList.empty())
					throw EraCMDException(CMDERROR_UNEXPECTED_COMMA);
				lastWordIsComma = true;
				continue;
			}

			// should i join with previous token?
			// text joining operator or 2ops operator joining text
			if (
				(token->type == Token::token_text && lastToken->type>Token::token_operators1_start && lastToken->type<Token::token_operators1_end) // operators before operand (join with text on the right)
				|| (token->type == Token::token_text && lastToken->type>Token::token_operators2_start && lastToken->type<Token::token_operators2_end) // operators with two operands (join with text on the right)
				|| (lastToken->type == Token::token_text && token->type>Token::token_operators2_start && token->type<Token::token_operators2_end) // operators with two operands (join with text on the left)
				|| (lastToken->type == Token::token_text && token->type>Token::token_operators3_start && token->type<Token::token_operators3_end) // operators after operand (join with text on the left)
				)
			{
				currentWord->tokenList.push_back(token);
				lastToken = token;
				continue;
			}

			// not joining, so currentWord is finished
			// add to list
			if (!lastWordIsComma) // add wordlist is finished, add it to elementlist
			{
				if (AddWordlistToElementList(currentWordList, outElements))
					currentWordList = std::make_shared<ElementWordList>(); // list was added, need a new one
			}
			lastWordIsComma = false;
		}
		else
		if (token->type == Token::token_comma)
		{
			throw EraCMDException(CMDERROR_UNEXPECTED_COMMA);
		}

		// start new word
		currentWord = std::make_shared<ElementWord>();
		currentWord->tokenList.push_back(token);
		currentWordList->wordList.push_back(currentWord);
		lastToken = token;
	}
	if (lastWordIsComma)
	{
		if (!collector)
			throw EraCMDException(CMDERROR_UNEXPECTED_COMMA);
	}
	else if (collector && ((lastToken->type>Token::token_operators1_start && lastToken->type<Token::token_operators1_end) || (lastToken->type>Token::token_operators2_start && lastToken->type<Token::token_operators2_end)))
	{
		currentWord->missingOperand = true;
		currentWordList->missingOperand = true;
	}

	AddWordlistToElementList(currentWordList, outElements);
	if (lastWordIsComma)
	{
		std::shared_ptr<Element> lastElement = *outElements.rbegin();
		lastElement->openedList = true;
	}

#ifdef _DEBUG
//	std::wstring dump = ElementArrayToDebugString(outElements, true);
//	std::wcout << dump << std::endl;
#endif
}

bool ScriptInterpreter::ProcessAndApplyInlineFlags(elementArray &elements)
{
	elementArray::const_iterator flagsBegin = elements.end();

	for (elementArray::const_reverse_iterator it = elements.rbegin(); it != elements.rend(); ++it)
	{
		const Element &element = **it;

		bool canBeInFlagsPart = (element.type == Element::element_word);
		if (! canBeInFlagsPart)
		{
			break;
		}

		bool isFlagBegin = gFlagsRegistry.IsFlag(element.GetString(), L"-");

		if (isFlagBegin)
		{
			flagsBegin = --(it.base());
		}
	}

	if (flagsBegin != elements.end())
	{
		elementArray::const_iterator it = flagsBegin;

		while (it != elements.end())
		{
			std::wstring flagName = (*it)->GetString();
			if (! gFlagsRegistry.IsFlag(flagName, L"-"))
			{
				throw EraCMDException(CMDERROR_FLAG_EXPECTED, flagName.c_str());
			}

			++it;
			std::wstring flagValue;
			if (it != elements.end())
			{
				std::wstring valueOrNextFlag = (*it)->GetString();
				bool isNextFlag = gFlagsRegistry.IsFlag(valueOrNextFlag, L"-");
				if (! isNextFlag)
				{
					flagValue = valueOrNextFlag;
					++it;
				}
			}

			gFlagsRegistry.GetFlag(flagName.substr(1))->SetTo(commandFlagsStack.CurrentFlags(), flagValue);
		}

		elements.erase(flagsBegin, elements.end());

		return true;
	}

	return false;
}

//////////////////////////////////////////////////////////////////////////
// CommandList
//////////////////////////////////////////////////////////////////////////

void CommandList::Add(std::wstring commandString, std::shared_ptr<Command> command)
{
	commandTable.push_back(std::make_pair(commandString, command));
}

std::shared_ptr<Command> CommandList::GetCommand(elementArray &elements /* could be more words, returns first matching command */, std::wstring &identifiedString)
{
	if (elements.empty()) return std::shared_ptr<Command>(); // prazdny command

	if (elements[0]->type != Element::element_word)
		throw EraCMDException(CMDERROR_INVALID_COMMAND_SYNTAX, elements[0]->GetString().c_str());

	for (size_t i = 0; i<commandTable.size(); ++i)
	{
		// split command into separate words
		std::vector<std::wstring> sequence;
		std::wstring tempS;
		std::wistringstream sequenceStream(commandTable[i].first);
		while (std::getline(sequenceStream, tempS, L' '))
		{
			sequence.push_back(tempS);
		}

		bool isMatchingSequence = true;
		std::wstring commandString;
		for (size_t cI = 0; cI < sequence.size(); ++cI)
		{
			if (cI >= elements.size() || elements[cI]->type != Element::element_word || _wcsicmp(elements[cI]->GetString().c_str(), sequence[cI].c_str()) != 0)
			{
				isMatchingSequence = false;
				break;
			}
			if (cI>0) commandString += L" ";
			commandString += elements[cI]->GetString();
		}
		if (isMatchingSequence)
		{
			identifiedString = commandString;
			elements.erase(elements.begin(), elements.begin() + sequence.size());
			return commandTable[i].second;
		}
	}

	return std::shared_ptr<Command>(); // NULL
}

std::shared_ptr<Command> CommandList::GetCommand(const std::wstring &commandString)
{
	for (size_t i = 0; i < commandTable.size(); ++i)
	{
		if (_wcsicmp(commandTable[i].first.c_str(), commandString.c_str()) == 0)
			return commandTable[i].second;
	}
	return std::shared_ptr<Command>(); // NULL
}

void CommandList::GetCommandStrings(std::vector<std::wstring> &outCommandList) const
{
	for (size_t i = 0; i < commandTable.size(); ++i)
		outCommandList.push_back(commandTable[i].first);
}

