/**
 * @file Command_PolicyRules.cpp
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "Command_PolicyRules.h"

#include <Common/era_api/err_codes.h>
#include <Common/scope_exit.h>
#include "../EracmdExceptions.h"

std::wstring Command_PolicyRules::GetServerName()
{
	Era_ServerInfo *serverInfo;

	uint32_t result_getServerInfo = era_GetServerInfo(&serverInfo);
	if (result_getServerInfo != 0)
	{
		throw EraCMDException_FromApi(result_getServerInfo);
	}
	scope_exit se_serverInfo(std::bind(era_ServerInfo_delete, serverInfo));

	return era_ServerInfo_GetServerName(serverInfo);
}

DWORD Command_PolicyRules::StrToDWORD(const std::wstring &str)
{
	std::wstringstream wss;
	wss << str;
	DWORD dwStr;
	wss >> dwStr;
	return dwStr;
}

void Command_PolicyRules::Execute(elementArray &elements, OutputHandler &out)
{
	Era_Query *query = era_Query_new();
	scope_exit se_reportQuery(std::bind(era_Query_delete, query));
	Era_TableReader *tableReader;
	uint32_t result_openQuery = era_OpenPolicyRulesQuery(&tableReader);
	if (result_openQuery != 0)
	{
		throw EraCMDException_FromApi(result_openQuery);
	}

	scope_exit se_tableReader(std::bind(era_TableReader_delete, tableReader));
	scope_exit_discard_return<DWORD> se_openQuery(std::bind(era_ClosePolicyRulesQuery, tableReader));

	int sortColumn1 = -1,
		sortColumn2 = -1;
	StringVector columns;
	uint32_t columnsCount = era_TableReader_GetColumnsCount(tableReader);
	columns.resize(columnsCount);
	for (uint32_t i = 0; i < columnsCount; ++i)
	{
		const Era_ColumnInfo *column = era_TableReader_GetColumn(tableReader, i);
		columns[i] = era_ColumnInfo_GetHeader(column);
		if (sortColumn1 == -1 && columns[i] == L"priority")
			sortColumn1 = i;
		if (sortColumn2 == -1 && columns[i] == L"primary_server")
			sortColumn2 = i;
	}

	StringVector prettyColumns;
	prettyColumns.resize(columnsCount);
	for (uint32_t i = 0; i < columnsCount; ++i)
	{
		const Era_ColumnInfo *column = era_TableReader_GetColumn(tableReader, i);
		prettyColumns[i] = era_ColumnInfo_GetPrettyHeader(column);
	}
	std::vector<StringVector> rows;
	StringVector row;

	bool readingComplete = false;
	do
	{
		Era_TableRow *tableRow;
		uint32_t result_readRow = era_ReadPolicyRulesRow(tableReader, &tableRow);
		if (result_readRow != 0)
		{
			throw EraCMDException_FromApi(result_readRow);
		}

		if (tableRow != NULL)
		{
			scope_exit se_tableRow(std::bind(era_TableRow_delete, tableRow));

			uint32_t prettyFieldsCount = era_TableRow_GetPrettyFieldsCount(tableRow);
			row.resize(prettyFieldsCount);
			for (uint32_t i = 0; i < prettyFieldsCount; ++i)
			{
				row[i] = era_TableRow_GetPrettyField(tableRow, i);
			}

			rows.push_back(std::move(row));
		}
		else
		{
			readingComplete = true;
		}
	} while (! readingComplete);

	std::wstring serverName = GetServerName();
	std::sort(
		rows.begin(),
		rows.end(),
		[serverName, sortColumn1, sortColumn2](StringVector &v1, StringVector &v2) -> bool
		{
			std::wstring primaryServer1 = v2[sortColumn2];
			std::wstring primaryServer2 = v1[sortColumn2];
			DWORD priority1 = StrToDWORD(v2[sortColumn1]);
			DWORD priority2 = StrToDWORD(v1[sortColumn1]);
			if (_wcsicmp(serverName.c_str(), primaryServer1.c_str()) == 0)
			{
				if (_wcsicmp(serverName.c_str(), primaryServer2.c_str()) == 0)
					return priority1 > priority2;
				else
					return true;
			}
			else
			{
				if (_wcsicmp(serverName.c_str(), primaryServer2.c_str()) == 0)
					return false;
				else
					return priority1 > priority2;
			}
		}
	);

	columns.resize(columns.size() - 3);
	prettyColumns.resize(prettyColumns.size() - 3);
	std::for_each(rows.begin(), rows.end(), [](StringVector &sv)
	{
		sv.resize(sv.size() - 3);
	});

	TableReader_Const tableReader2(columns, prettyColumns, rows);
	out.Print(tableReader2);
}

