/**
 * @file Command_Show.cpp
 * @author Copyright (c) 2014 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "Command_Show.h"

#include <Common/era_api/era_api.h>
#include <Common/era_api/err_codes.h>
#include <Common/scope_exit.h>
#include "../util/stringUtils.h"
#include "../Element.h"
#include "../EracmdExceptions.h"
#include "../TableReader.h"
#include "../Token.h"

Command_Show::Command_Show(CommandFlagsStack &commandFlagsStack)
:
	commandFlagsStack(commandFlagsStack)
{
}

void Command_Show::Execute(elementArray &elements, OutputHandler &out)
{
	Era_Query *query = era_Query_new();
	scope_exit se_query(std::bind(era_Query_delete, query));

	bool isCountOnly = false;

	era_Query_SetTable(query, GetArgument(0)->GetString().c_str());
	if (std::shared_ptr<Element> element = GetArgument(1))
	{
		SetColumns(*element, *query, isCountOnly);
	}
	if (std::shared_ptr<Element> element = GetArgument(2))
	{
		SetConditions(*element, *query);
	}
	if (std::shared_ptr<Element> element = GetArgument(3))
	{
		SetGroupBy(*element, *query);
	}
	if (std::shared_ptr<Element> element = GetArgument(4))
	{
		SetOrderBy(*element, *query);
	}
	if (std::shared_ptr<Element> element = GetArgument(5))
	{
		SetSkip(*element, *query);
	}
	if (std::shared_ptr<Element> element = GetArgument(6))
	{
		SetLimit(*element, *query);
	}

	if (isCountOnly) {
		DoCount(*query, out);
	} else {
		DoData(*query, out);
	}
}

void Command_Show::SetColumns(const Element &element, Era_Query &query, bool & isCountOnly)
{
	isCountOnly = false;

	std::vector<const ElementWord *> words = element.GetWords();
	if (words.size() == 1) {
		if (::_wcsicmp(words[0]->GetString().c_str(), L"*") == 0) {
			era_Query_SetAllColumns(&query);
			return;
		}
		if (::_wcsicmp(words[0]->GetString().c_str(), L"count") == 0) {
			isCountOnly = true;
			return;
		}
	}

	for (std::vector<const ElementWord *>::const_iterator it = words.begin(); it != words.end(); ++it)
	{
		std::wstring column = (*it)->GetString();

		era_Query_AddColumn(&query, column.c_str());
	}
}

void Command_Show::SetConditions(const Element &element, Era_Query &query)
{
	std::vector<const ElementWord *> conditions = element.GetWords();
	for (std::vector<const ElementWord *>::const_iterator it = conditions.begin(); it != conditions.end(); ++it)
	{
		const ElementWord &condition = **it;

		if (condition.tokenList.size() < 2)
		{
			throw EraCMDException(CMDERROR_INVALID_CONDITION_SYNTAX, condition.GetString().c_str());
		}

		std::wstring column = condition.tokenList[0]->text;
		Token::TokenType signToken = condition.tokenList[1]->type;

		if (signToken == Token::token_in)
		{
			Era_Sign sign = ERA_SIGN__EQUALS;

			bool first = true;
			for (size_t i = 2; i < condition.tokenList.size(); ++i)
			{
				std::shared_ptr<Token> token = condition.tokenList[i];

				std::wstring value = token->text;
				if (first)
				{
					era_Query_AddCondition(&query, column.c_str(), sign, value.c_str());
				}
				else
				{
					era_Query_AddConditionAlternative(&query, column.c_str(), sign, value.c_str());
				}

				if (first)
				{
					first = false;
				}
			}
		}
		else if (signToken == Token::token_notin)
		{
			Era_Sign sign = ERA_SIGN__NOT_EQUALS;

			for (size_t i = 2; i < condition.tokenList.size(); ++i)
			{
				std::shared_ptr<Token> token = condition.tokenList[i];

				std::wstring value = token->text;

				era_Query_AddCondition(&query, column.c_str(), sign, value.c_str());
			}
		}
		else
		{
			if (condition.tokenList.size() != 3)
			{
				throw EraCMDException(CMDERROR_INVALID_CONDITION_SYNTAX, condition.GetString().c_str());
			}

			std::wstring column = condition.tokenList[0]->text;

			Era_Sign sign;
			switch (signToken)
			{
				case Token::token_equals:
					sign = ERA_SIGN__EQUALS;
					break;
				case Token::token_notequals:
					sign = ERA_SIGN__NOT_EQUALS;
					break;
				case Token::token_greater:
					sign = ERA_SIGN__GREATER;
					break;
				case Token::token_less:
					sign = ERA_SIGN__LESS;
					break;
				case Token::token_greaterequal:
					sign = ERA_SIGN__GREATER_EQUAL;
					break;
				case Token::token_lessequal:
					sign = ERA_SIGN__LESS_EQUAL;
					break;
				case Token::token_like:
					sign = ERA_SIGN__LIKE;
					break;
				case Token::token_notlike:
					sign = ERA_SIGN__NOTLIKE;
					break;
				default:
					throw std::exception("unsupported sign");
			};

			std::wstring value = condition.tokenList[2]->text;

			era_Query_AddCondition(&query, column.c_str(), sign, value.c_str());
		}
	}
}

void Command_Show::SetGroupBy(const Element &element, Era_Query &query)
{
	std::vector<const ElementWord *> columns = element.GetWords();
	for (std::vector<const ElementWord *>::const_iterator it = columns.begin(); it != columns.end(); ++it)
	{
		std::wstring column = (*it)->GetString();

		era_Query_AddGroupBy(&query, column.c_str());
	}
}

void Command_Show::SetOrderBy(const Element &element, Era_Query &query)
{
	std::vector<const ElementWord *> words = element.GetWords();
	for (std::vector<const ElementWord *>::const_iterator it = words.begin(); it != words.end(); ++it)
	{
		const ElementWord &w = **it;

		std::wstring column = w.tokenList[0]->text;

		bool desc = false;
		if (w.tokenList.size() > 1)
		{
			std::wstring ascDesc = w.tokenList[1]->text;

			if (::_wcsicmp(ascDesc.c_str(), L"-asc") == 0)
			{
				desc = false;
			}
			else if (::_wcsicmp(ascDesc.c_str(), L"-desc") == 0)
			{
				desc = true;
			}
			else
			{
				throw EraCMDException(CMDERROR_ASC_OR_DESC_EXPECTED_IN_ORDER, ascDesc.c_str());
			}
		}

		era_Query_AddOrderBy(&query, column.c_str(), desc);
	}
}

void Command_Show::SetSkip(const Element &element, Era_Query &query)
{
	int num = 0;
	try
	{
		num = WstringToUint(element.GetString());
	}
	catch (...)
	{
		throw EraCMDException(CMDERROR_NUMBER_EXPECTED_IN_SKIP, element.GetString().c_str());
	}

	if (num < 0)
	{
		throw EraCMDException(CMDERROR_NON_NEGATIVE_NUMBER_EXPECTED_IN_SKIP, element.GetString().c_str());
	}

	era_Query_SetSkip(&query, num);
}

void Command_Show::SetLimit(const Element &element, Era_Query &query)
{
	int num = 0;
	try
	{
		num = WstringToUint(element.GetString());
	}
	catch (...)
	{
		throw EraCMDException(CMDERROR_NUMBER_EXPECTED_IN_LIMIT, element.GetString().c_str());
	}

	if (num < 0)
	{
		throw EraCMDException(CMDERROR_NON_NEGATIVE_NUMBER_EXPECTED_IN_LIMIT, element.GetString().c_str());
	}

	era_Query_SetLimit(&query, num);
}

void Command_Show::DoData(Era_Query &query, OutputHandler &out)
{
	Era_TableReader *tableReader;
	uint32_t result_openQuery = era_OpenQuery(&query, &tableReader);
	if (result_openQuery != 0)
	{
		throw EraCMDException_FromApi(result_openQuery);
	}

	scope_exit se_tableReader(std::bind(era_TableReader_delete, tableReader));
	scope_exit_discard_return<DWORD> se_openQuery(std::bind(era_CloseQuery, tableReader));

	StringVector columns;
	uint32_t columnsCount = era_TableReader_GetColumnsCount(tableReader);
	columns.resize(columnsCount);
	for (uint32_t i = 0; i < columnsCount; ++i)
	{
		const Era_ColumnInfo *columnInfo = era_TableReader_GetColumn(tableReader, i);
		columns[i] = era_ColumnInfo_GetHeader(columnInfo);
	}

	StringVector prettyColumns;
	prettyColumns.resize(columnsCount);
	for (uint32_t i = 0; i < columnsCount; ++i)
	{
		const Era_ColumnInfo *columnInfo = era_TableReader_GetColumn(tableReader, i);
		prettyColumns[i] = era_ColumnInfo_GetPrettyHeader(columnInfo);
	}

	TableReader_Api tableReader2(columns, prettyColumns, *tableReader, commandFlagsStack.CurrentFlags().fieldFormatting == CommandFlags::FIELD_PRETTY);
	out.Print(tableReader2);
}

void Command_Show::DoCount(Era_Query &query, OutputHandler &out)
{
	uint32_t count = 0;
	uint32_t result_getCount = era_GetCount(&query, &count);
	if (result_getCount != 0)
	{
		throw EraCMDException_FromApi(result_getCount);
	}

	StringVector header(1);
	header[0] = L"count";

	StringVector prettyHeader(1);
	prettyHeader[0] = LoadStringFromResource(IDS_COUNT);

	std::vector<StringVector> rows;
	StringVector row(1);
	row[0] = std::to_wstring((unsigned long long)count);
	rows.push_back(std::move(row));

	TableReader_Const reader(header, prettyHeader, rows);
	out.Print(reader);
}

