/**
 * @file Eracmd.cpp
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"

#include <Common/era_api/era_api.h>
#include <Common/scope_exit.h>
#include "autocomplete/Autocomplete.h"
#include "console_colors/ConsoleColors_Win.h"
#include "flags/CommandFlagsStack.h"
#include "flags/FlagsRegistry.h"
#include "output_format/TextWidth_Console.h"
#include "output_format/TextWidth_File.h"
#include "util/fileUtils.h"
#include "util/stringUtils.h"
#include "ConsoleHistory.h"
#include "ConsoleKeyboardInput.h"
#include "EracmdExceptions.h"
#include "ErrorMessages.h"
#include "FileLoader.h"
#include "resource.h"
#include "ScriptInterpreter.h"
#include "Version2.h"

ConsoleColors_Win gConsoleColors;
TextWidth_Console gTextWidth_Console;
TextWidth_File gTextWidth_File;
FlagsRegistry gFlagsRegistry;

void ShellMode(ScriptInterpreter &scriptInterpreter, bool &partiallyFailed);
void RunStartupScript(ScriptInterpreter &scriptInterpreter, bool &partiallyFailed);

DWORD ExecuteCommands(ScriptInterpreter &scriptInterpreter, const std::wstring &commandLine, bool connect, bool &outSIResult, bool &partiallyFailed)
{
	DWORD errCode = 0;
	try
	{
		outSIResult = scriptInterpreter.ExecuteScript(commandLine, connect, NULL);
		if (partiallyFailed)
		{
			// std::wcerr << gConsoleColors.LIGHT_RED << LoadStringFromResource(IDS_TXPLN_SOME_ERRORS_IGNORED) << " " << gConsoleColors.DEFAULT << std::endl;
			errCode = 7;
		}
	}
	catch (const EraCMDException &e)
	{
		e.Print(std::wcerr);
		errCode = e.errorCode;
		if (typeid(e) == typeid(const EraCMDException_Pause) || typeid(e) == typeid(const EraCMDException_FromApi_Pause))
		{
			std::wcout << LoadStringFromResource(IDS_PRESS_ENTER);
			std::wcin.ignore();
		}
	}
	catch (const std::exception &e)
	{
		std::wcerr << gConsoleColors.LIGHT_RED << LoadStringFromResource(IDS_ERROR_UPPERCASE) << " " << gConsoleColors.DEFAULT << e.what() << std::endl;
		errCode = -4;
	}
	return errCode;
}


BOOL WINAPI CtrlHandler(DWORD dwCtrlType)
{
	if (dwCtrlType == CTRL_C_EVENT)
	{
		gConsoleColors.SetOrigColors();
	}

	return FALSE;
}

int wmain(int argc, WCHAR *argv[])
{
	_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);

	::SetConsoleCtrlHandler(CtrlHandler, TRUE);

	_setmode(_fileno(stdout), _O_U16TEXT);
	_setmode(_fileno(stderr), _O_U16TEXT);

	scope_exit se(era_ThreadDataCleanup);

	// process commandline input
	std::wstring commandLine = GetCommandLineW();

	// remove argv[0] from commandline
	std::wstring argv0 = argv[0];
	size_t argv0Len = argv0.length();
	if (argv0[0] != L'"' && commandLine[0] == L'"') argv0Len += 2;
	commandLine.erase(0, argv0Len);

	// initialize error table
	LoadErrorMessages(GetModuleHandle(NULL));

	bool connectToServer = true;

	// help
	if (argc == 2)
	{
		// check prefix
		bool hasPrefix = true;
		std::wstring arg(argv[1]);
		if (arg.substr(0, 2) == L"--")
			arg.erase(0, 2);
		else
		if ((arg.size() > 0) && (arg[0] == L'/' || arg[0] == L'/' || arg[0] == L'-'))
			arg.erase(0, 1);
		else hasPrefix = false;

		// check the rest
		if (hasPrefix && ((_wcsicmp(arg.c_str(), L"help") == 0) || (_wcsicmp(arg.c_str(), L"h") == 0) || (_wcsicmp(arg.c_str(), L"?") == 0)))
		{
			connectToServer = false;
			commandLine = L"HELP -paged";
		}
	}

	// execute commandline commands or just connect and switch to shell mode
	CommandFlagsStack commandFlagsStack;
	bool partiallyFailed = false;

	ScriptInterpreter si(commandFlagsStack, partiallyFailed, L"");
	bool commandExecuted = false;
	DWORD result = ExecuteCommands(si, commandLine, connectToServer, commandExecuted, partiallyFailed);

	if (result != 0)
		return result;

	if (commandExecuted == false)
		ShellMode(si, partiallyFailed);

	return 0;
}

void ShellMode(ScriptInterpreter &scriptInterpreter, bool &partiallyFailed)
{
	scriptInterpreter.GetCommandFlagsStack().CurrentFlags().outputFormatting = CommandFlags::FORMAT_TABLE;

	// welcome

	std::wstring histFile = GetProgramDataFilePath(L"eracmd_history.txt");
	gCmdHistory.FileName = histFile;

	if (FileExists(histFile.c_str()))
	{
		gCmdHistory.Active = true;
		gCmdHistory.Load();
	}

	// cursor position
	HANDLE hOut = GetStdHandle(STD_OUTPUT_HANDLE);
	CONSOLE_SCREEN_BUFFER_INFO SBInfo;
	GetConsoleScreenBufferInfo(hOut, &SBInfo);
	int cursorX = SBInfo.dwCursorPosition.X;
	int cursorY = SBInfo.dwCursorPosition.Y - SBInfo.srWindow.Top;
	if (SBInfo.srWindow.Top) cursorY--;

	// font size
	CONSOLE_FONT_INFO fontInfo;
	GetCurrentConsoleFont(hOut, FALSE, &fontInfo);
	int fontX = fontInfo.dwFontSize.X;
	int fontY = fontInfo.dwFontSize.Y;

	// logo + version
	WCHAR wchVersionString[128] = L"";
	GetVersionString(RA_CMD_VERSION, wchVersionString, sizeof(wchVersionString) / sizeof(wchVersionString[0]));

	std::wstring welcomeMessage = LoadStringFromResource(IDS_WELCOME_MESSAGE);

	unsigned char esetL24[] = { 0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x6 ,0x34 ,0x5D ,0x6A ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x6B ,0x64 ,0x42 ,0x12 ,0x0 ,0x2 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x9 ,0x73 ,0xD5 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xE5 ,0x97 ,0x20 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x1C ,0xC4 ,0xFF ,0xFF ,0xFA ,0xCE ,0xAB ,0xA1 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA1 ,0xAE ,0xFB ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xE7 ,0x46 ,0x1 ,0x0 ,0x0 ,0x14 ,0xD3 ,0xFF ,0xFC ,0x95 ,0x2A ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x16 ,0xF4 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xF7 ,0x3F ,0x0 ,0x0 ,0x9C ,0xFF ,0xF5 ,0x4D ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x20 ,0xF5 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xDB ,0xB ,0x1C ,0xF4 ,0xFF ,0x73 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x20 ,0xF5 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0x5E ,0x67 ,0xFF ,0xDD ,0x7 ,0x0 ,0x0 ,0x0 ,0x0 ,0x5 ,0x2E ,0x4B ,0x5B ,0x57 ,0x43 ,0x21 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x2B ,0x4A ,0x58 ,0x5F ,0x57 ,0x44 ,0x20 ,0x0 ,0x0 ,0x0 ,0x20 ,0xF5 ,0xFF ,0xFF ,0xFF ,0xFB ,0xCE ,0xAE ,0x9C ,0x9F ,0xB5 ,0xD8 ,0xFF ,0xFF ,0xFF ,0xFF ,0xBD ,0xA9 ,0xA9 ,0xA9 ,0xA9 ,0xA9 ,0xA9 ,0xA8 ,0xCE ,0xFF ,0xFF ,0xFF ,0xFF ,0xAC ,0x9D ,0xFF ,0x8F ,0x0 ,0x0 ,0x0 ,0x0 ,0x2E ,0xCB ,0xFD ,0xFF ,0xFF ,0xFF ,0xFF ,0xF4 ,0x95 ,0x5 ,0x0 ,0x0 ,0xF ,0xB2 ,0xFB ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xF2 ,0x83 ,0x1 ,0x0 ,0x20 ,0xF5 ,0xFF ,0xFF ,0xD9 ,0x37 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x7 ,0x5C ,0xF6 ,0xFF ,0xFF ,0x32 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x67 ,0xFF ,0xFF ,0xFF ,0xFF ,0xE3 ,0xBF ,0xFF ,0x4C ,0x0 ,0x0 ,0x0 ,0x7 ,0xD2 ,0xFF ,0xFF ,0xFF ,0xEF ,0xF4 ,0xFF ,0xFF ,0xFF ,0x71 ,0x0 ,0x0 ,0x75 ,0xFF ,0xFF ,0xFB ,0xD1 ,0xC5 ,0xE9 ,0xFF ,0xFF ,0xFF ,0x32 ,0x0 ,0x21 ,0xF5 ,0xFF ,0xFB ,0x3C ,0x0 ,0x0 ,0x3 ,0x29 ,0x24 ,0x0 ,0x0 ,0x0 ,0x7C ,0xFF ,0xFF ,0x41 ,0x5 ,0x3 ,0x0 ,0x0 ,0x0 ,0x4 ,0x5 ,0x72 ,0xFF ,0xFF ,0xFF ,0xFF ,0xEF ,0xD9 ,0xFE ,0x28 ,0x0 ,0x0 ,0x0 ,0x38 ,0xFF ,0xFF ,0xFD ,0x61 ,0xD ,0x15 ,0xA9 ,0xFF ,0xFF ,0xD1 ,0x3 ,0x0 ,0xB0 ,0xFF ,0xFF ,0xA6 ,0x0 ,0x0 ,0x19 ,0xA1 ,0xAC ,0xB1 ,0x34 ,0x0 ,0x21 ,0xF5 ,0xFF ,0xD1 ,0x0 ,0x0 ,0x3 ,0xB1 ,0xFF ,0xFF ,0x77 ,0x0 ,0x0 ,0x29 ,0xFB ,0xFF ,0xE0 ,0xD7 ,0x94 ,0x0 ,0x0 ,0x8 ,0xBA ,0xD7 ,0xE8 ,0xFF ,0xFF ,0xFF ,0xFF ,0xED ,0xEB ,0xF5 ,0x15 ,0x0 ,0x0 ,0x0 ,0x60 ,0xFF ,0xFF ,0xF3 ,0x36 ,0x26 ,0x24 ,0x7B ,0xFF ,0xFF ,0xEE ,0x13 ,0x0 ,0xB7 ,0xFF ,0xFF ,0xD5 ,0x57 ,0x3C ,0x31 ,0x20 ,0x8 ,0x0 ,0x0 ,0x0 ,0x21 ,0xF5 ,0xFF ,0xB1 ,0x0 ,0x0 ,0xA ,0xC0 ,0xD2 ,0xD3 ,0x96 ,0x0 ,0x0 ,0x14 ,0xEE ,0xFF ,0xFF ,0xFF ,0xB6 ,0x0 ,0x0 ,0x9 ,0xE4 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xED ,0xF2 ,0xEE ,0x10 ,0x0 ,0x0 ,0x0 ,0x70 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0x1F ,0x0 ,0x84 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFA ,0xE8 ,0xAA ,0x16 ,0x0 ,0x21 ,0xF5 ,0xFF ,0xA7 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0xE3 ,0xFF ,0xFF ,0xFF ,0xB0 ,0x0 ,0x0 ,0x9 ,0xDD ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xED ,0xF2 ,0xED ,0x10 ,0x0 ,0x0 ,0x0 ,0x72 ,0xFF ,0xFF ,0xF8 ,0xB3 ,0xAF ,0xAF ,0xAD ,0xA8 ,0xA8 ,0xA1 ,0x14 ,0x0 ,0x14 ,0xA7 ,0xEE ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0x90 ,0x0 ,0x20 ,0xF5 ,0xFF ,0xA6 ,0x0 ,0x0 ,0x7 ,0x66 ,0x6F ,0x6F ,0x71 ,0x77 ,0x77 ,0x7D ,0xEF ,0xFF ,0xFF ,0xFF ,0xB0 ,0x0 ,0x0 ,0x9 ,0xDD ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xED ,0xED ,0xF3 ,0x13 ,0x0 ,0x0 ,0x0 ,0x66 ,0xFF ,0xFF ,0xEF ,0x8 ,0x0 ,0x0 ,0x0 ,0x2 ,0x2 ,0x2 ,0x1 ,0x0 ,0x0 ,0x0 ,0xF ,0x32 ,0x4C ,0x52 ,0x62 ,0xD5 ,0xFF ,0xFF ,0xBE ,0x0 ,0x1F ,0xF5 ,0xFF ,0xAF ,0x0 ,0x0 ,0xD ,0xF6 ,0xFF ,0xFF ,0xFF ,0xF8 ,0xF8 ,0xF7 ,0xFD ,0xFF ,0xFF ,0xFF ,0xAF ,0x0 ,0x0 ,0x9 ,0xDD ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xED ,0xDD ,0xFD ,0x24 ,0x0 ,0x0 ,0x0 ,0x42 ,0xFF ,0xFF ,0xFD ,0x4A ,0x0 ,0x0 ,0x6B ,0xE6 ,0xE6 ,0xDD ,0x18 ,0x0 ,0x7F ,0xB3 ,0xAE ,0x82 ,0x0 ,0x0 ,0x0 ,0x8A ,0xFF ,0xFF ,0xBD ,0x0 ,0x1F ,0xF5 ,0xFF ,0xCF ,0x0 ,0x0 ,0x1 ,0xAB ,0xFF ,0xFF ,0x9F ,0x11 ,0x11 ,0x1E ,0xEB ,0xFF ,0xFF ,0xFF ,0xAF ,0x0 ,0x0 ,0x9 ,0xDD ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xED ,0xC4 ,0xFF ,0x43 ,0x0 ,0x0 ,0x0 ,0xD ,0xE5 ,0xFF ,0xFF ,0xEF ,0xBF ,0xC2 ,0xF7 ,0xFF ,0xFF ,0xD8 ,0x6 ,0x0 ,0x98 ,0xFF ,0xFF ,0xFF ,0xAB ,0x8A ,0x97 ,0xE5 ,0xFF ,0xFF ,0x8E ,0x0 ,0x20 ,0xF5 ,0xFF ,0xFA ,0x36 ,0x0 ,0x0 ,0xA ,0x3B ,0x3A ,0x9 ,0x0 ,0x0 ,0x2C ,0xFB ,0xFF ,0xFF ,0xFF ,0xAF ,0x0 ,0x0 ,0x9 ,0xDD ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xEE ,0xA4 ,0xFF ,0x80 ,0x0 ,0x0 ,0x0 ,0x0 ,0x53 ,0xEF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xEC ,0x45 ,0x0 ,0x0 ,0x2C ,0xE6 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xC6 ,0x18 ,0x0 ,0x20 ,0xF5 ,0xFF ,0xFF ,0xCE ,0x29 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x22 ,0xC3 ,0xFF ,0xFF ,0xFF ,0xFF ,0xAA ,0x0 ,0x0 ,0x0 ,0xDB ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xE9 ,0x72 ,0xFF ,0xD1 ,0x1 ,0x0 ,0x0 ,0x0 ,0x0 ,0x21 ,0x6D ,0x95 ,0xA2 ,0xA0 ,0x93 ,0x6A ,0x1E ,0x0 ,0x0 ,0x0 ,0x0 ,0x1D ,0x70 ,0x97 ,0xA2 ,0xA4 ,0x9D ,0x89 ,0x4E ,0x6 ,0x0 ,0x0 ,0x20 ,0xF5 ,0xFF ,0xFF ,0xFF ,0xF4 ,0xB6 ,0x8A ,0x77 ,0x77 ,0x8B ,0xB3 ,0xEF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xD6 ,0x7B ,0x7B ,0x80 ,0xED ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xB6 ,0x28 ,0xFB ,0xFF ,0x55 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x20 ,0xF5 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0x70 ,0x0 ,0xB3 ,0xFF ,0xE6 ,0x2F ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x20 ,0xF5 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xEA ,0x15 ,0x0 ,0x24 ,0xEB ,0xFF ,0xEB ,0x6D ,0xC ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x12 ,0xF4 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0x5C ,0x0 ,0x0 ,0x0 ,0x39 ,0xE1 ,0xFF ,0xFF ,0xDE ,0xA5 ,0x78 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x71 ,0x83 ,0xF9 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFA ,0x6E ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x1B ,0x9B ,0xEF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFC ,0xBA ,0x3E ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x2B ,0x6D ,0x91 ,0x9C ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9B ,0x9C ,0x97 ,0x7B ,0x41 ,0x2 ,0x0 ,0x0 ,0x0 ,0x0};
	unsigned char esetL20[] = { 0x0 ,0x0 ,0x0 ,0x3 ,0x10 ,0x44 ,0x6D ,0x75 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x74 ,0x71 ,0x51 ,0x1B ,0x5 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x15 ,0x7F ,0xEB ,0xFF ,0xFF ,0xF7 ,0xF6 ,0xF6 ,0xF6 ,0xF6 ,0xF6 ,0xF6 ,0xF6 ,0xF6 ,0xF6 ,0xF6 ,0xF6 ,0xF6 ,0xF6 ,0xF6 ,0xF6 ,0xF6 ,0xF7 ,0xF6 ,0xFC ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xF6 ,0x9E ,0x2A ,0x0 ,0x0 ,0x0 ,0x1B ,0xB5 ,0xFF ,0xDD ,0x8F ,0x5C ,0x4C ,0x4A ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4A ,0x74 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFE ,0xFF ,0xFF ,0xD8 ,0x38 ,0x0 ,0x4 ,0x9D ,0xFF ,0xAA ,0x2A ,0x2 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x3A ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFE ,0xFF ,0xFF ,0xD3 ,0xC ,0x24 ,0xFF ,0xCA ,0x21 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x41 ,0xFF ,0xFF ,0xFE ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFE ,0xFF ,0xFE ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFE ,0xFF ,0xFF ,0x60 ,0x76 ,0xFF ,0x6C ,0x0 ,0x0 ,0x0 ,0x4 ,0x25 ,0x51 ,0x68 ,0x64 ,0x4A ,0x1D ,0x1 ,0x0 ,0x0 ,0x1 ,0x1D ,0x4F ,0x64 ,0x6D ,0x62 ,0x45 ,0x14 ,0x0 ,0x0 ,0x41 ,0xFF ,0xFF ,0xFE ,0xEF ,0xB8 ,0x97 ,0x8D ,0xA0 ,0xC7 ,0xF8 ,0xFF ,0xFF ,0xC3 ,0x9B ,0x9A ,0x9A ,0x9A ,0x9A ,0x98 ,0xCB ,0xFF ,0xFF ,0xFF ,0xB7 ,0xAE ,0xE7 ,0x35 ,0x0 ,0x0 ,0x4 ,0x80 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xED ,0x4F ,0x1 ,0x0 ,0x4A ,0xF6 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xDA ,0x25 ,0x0 ,0x41 ,0xFF ,0xFF ,0xC6 ,0x1C ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x4D ,0xF0 ,0xFF ,0x57 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x70 ,0xFF ,0xFF ,0xFF ,0xEB ,0xD0 ,0xD2 ,0xF ,0x0 ,0x0 ,0x28 ,0xFF ,0xFF ,0xCD ,0x7F ,0x85 ,0xE5 ,0xFF ,0xD0 ,0x11 ,0x3 ,0xB7 ,0xFF ,0xE6 ,0x6D ,0x5B ,0x8F ,0xE1 ,0xF4 ,0x57 ,0x0 ,0x42 ,0xFF ,0xFF ,0x38 ,0x0 ,0xB ,0x75 ,0xA1 ,0x4E ,0x0 ,0x1 ,0x8A ,0xFF ,0xA9 ,0x6D ,0x40 ,0x0 ,0x0 ,0x47 ,0x69 ,0xB5 ,0xFF ,0xFF ,0xFF ,0xEF ,0xE8 ,0xC1 ,0x0 ,0x0 ,0x0 ,0x61 ,0xFF ,0xFF ,0x6A ,0xA ,0xC ,0x9E ,0xFF ,0xFF ,0x26 ,0x3 ,0xD4 ,0xFF ,0xE2 ,0x41 ,0x17 ,0x19 ,0x2A ,0x1F ,0xF ,0x0 ,0x42 ,0xFF ,0xE1 ,0x1B ,0x0 ,0x22 ,0xEA ,0xFC ,0xB8 ,0x3 ,0x0 ,0x5F ,0xFF ,0xFC ,0xFD ,0x9D ,0x0 ,0x0 ,0xAC ,0xFA ,0xFD ,0xFF ,0xFF ,0xFF ,0xEE ,0xF2 ,0xB6 ,0x0 ,0x0 ,0x0 ,0x79 ,0xFF ,0xFF ,0xF7 ,0xF4 ,0xF4 ,0xFB ,0xFF ,0xFF ,0x39 ,0x2 ,0xAB ,0xFF ,0xFE ,0xF5 ,0xF2 ,0xF1 ,0xE6 ,0xBA ,0x24 ,0x0 ,0x42 ,0xFF ,0xD6 ,0x1A ,0x0 ,0x1 ,0x8 ,0x9 ,0x6 ,0x0 ,0x0 ,0x44 ,0xFF ,0xFF ,0xFF ,0x9B ,0x0 ,0x0 ,0xAA ,0xFE ,0xFF ,0xFE ,0xFF ,0xFF ,0xEE ,0xF2 ,0xB5 ,0x0 ,0x0 ,0x0 ,0x7C ,0xFF ,0xFF ,0xBD ,0xA0 ,0xA0 ,0x9C ,0x98 ,0x9D ,0x21 ,0x0 ,0x27 ,0xBD ,0xED ,0xF3 ,0xF4 ,0xF6 ,0xFE ,0xFF ,0xB1 ,0x0 ,0x41 ,0xFF ,0xD5 ,0x1A ,0x0 ,0x16 ,0x7E ,0x7D ,0x7E ,0x84 ,0x84 ,0xAB ,0xFE ,0xFF ,0xFF ,0x9B ,0x0 ,0x0 ,0xAA ,0xFD ,0xFF ,0xFF ,0xFF ,0xFF ,0xEE ,0xEB ,0xBF ,0x0 ,0x0 ,0x0 ,0x68 ,0xFF ,0xFF ,0x4E ,0x0 ,0x0 ,0x14 ,0x32 ,0x32 ,0xA ,0x0 ,0x20 ,0x21 ,0x2F ,0x27 ,0x2B ,0x4E ,0xDF ,0xFF ,0xD7 ,0x0 ,0x40 ,0xFF ,0xE0 ,0x1B ,0x0 ,0x28 ,0xFF ,0xFF ,0xFD ,0xC7 ,0xC6 ,0xD7 ,0xFF ,0xFF ,0xFF ,0x9A ,0x0 ,0x0 ,0xAA ,0xFD ,0xFF ,0xFF ,0xFF ,0xFF ,0xEE ,0xD5 ,0xCF ,0xB ,0x0 ,0x0 ,0x31 ,0xFF ,0xFF ,0xC0 ,0x61 ,0x60 ,0xCC ,0xFF ,0xFF ,0x2A ,0x3 ,0xB3 ,0xEA ,0xD7 ,0x5C ,0x38 ,0x51 ,0xD7 ,0xFF ,0xC4 ,0x0 ,0x40 ,0xFF ,0xFF ,0x36 ,0x0 ,0xA ,0x76 ,0xA9 ,0x63 ,0x0 ,0x0 ,0x57 ,0xFF ,0xFF ,0xFF ,0x9A ,0x0 ,0x0 ,0xAA ,0xFD ,0xFF ,0xFF ,0xFF ,0xFF ,0xEE ,0xB4 ,0xE3 ,0x2D ,0x0 ,0x0 ,0x7 ,0xA5 ,0xFF ,0xFF ,0xFC ,0xFC ,0xFF ,0xFF ,0xA5 ,0xB ,0x1 ,0x72 ,0xFF ,0xFF ,0xF8 ,0xEB ,0xF3 ,0xFF ,0xFE ,0x53 ,0x0 ,0x41 ,0xFF ,0xFF ,0xBD ,0x11 ,0x0 ,0x0 ,0x1 ,0x0 ,0x0 ,0x19 ,0xC3 ,0xFF ,0xFF ,0xFF ,0x92 ,0x0 ,0x0 ,0xA3 ,0xFD ,0xFF ,0xFF ,0xFF ,0xFF ,0xEF ,0x80 ,0xFF ,0x62 ,0x0 ,0x0 ,0x0 ,0xC ,0x52 ,0x98 ,0xAE ,0xAC ,0x98 ,0x56 ,0xD ,0x0 ,0x0 ,0x8 ,0x4E ,0x99 ,0xAD ,0xB0 ,0xA7 ,0x81 ,0x2C ,0x2 ,0x0 ,0x41 ,0xFF ,0xFF ,0xFE ,0xE7 ,0x9E ,0x72 ,0x67 ,0x76 ,0xA2 ,0xE7 ,0xFE ,0xFF ,0xFE ,0xFF ,0xC5 ,0x66 ,0x68 ,0xCE ,0xFE ,0xFF ,0xFF ,0xFF ,0xFF ,0xC0 ,0x2F ,0xFF ,0xB9 ,0x17 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x41 ,0xFF ,0xFF ,0xFE ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFE ,0xFF ,0xFF ,0xFE ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0x70 ,0x5 ,0xB2 ,0xFF ,0x90 ,0x1D ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x3B ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFE ,0xFE ,0xFF ,0xE4 ,0x14 ,0x0 ,0x29 ,0xD0 ,0xFF ,0xC6 ,0x6F ,0x44 ,0x30 ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x5D ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFE ,0xFF ,0xFF ,0xEB ,0x4B ,0x0 ,0x0 ,0x0 ,0x24 ,0x9E ,0xFF ,0xFF ,0xF0 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xE5 ,0xEE ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xBC ,0x3F ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x3 ,0x31 ,0x7A ,0x9D ,0xA3 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA2 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA1 ,0xA2 ,0xA0 ,0x86 ,0x45 ,0x6 ,0x0 ,0x0 ,0x0};
	unsigned char esetL18[] = { 0x0 ,0x0 ,0x1 ,0x23 ,0x4F ,0x7D ,0x8F ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x90 ,0x83 ,0x5B ,0x2C ,0x7 ,0x0 ,0x0 ,0x0 ,0x7 ,0x50 ,0xC3 ,0xF3 ,0xED ,0xD4 ,0xD0 ,0xD1 ,0xD1 ,0xD1 ,0xD1 ,0xD1 ,0xD1 ,0xD1 ,0xD1 ,0xD1 ,0xD1 ,0xD1 ,0xD1 ,0xD1 ,0xD1 ,0xD1 ,0xD4 ,0xF8 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xF8 ,0xD2 ,0x69 ,0x11 ,0x0 ,0x4 ,0x59 ,0xDC ,0xEB ,0x89 ,0x40 ,0x2B ,0x28 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x35 ,0xE0 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xEC ,0x7A ,0xB ,0x31 ,0xD8 ,0xE0 ,0x2F ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x14 ,0xDC ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xF1 ,0x4E ,0xA1 ,0xDB ,0x33 ,0x0 ,0x0 ,0x11 ,0x6F ,0x9D ,0xAB ,0xA9 ,0x97 ,0x5F ,0x7 ,0x0 ,0x5 ,0x5C ,0x9A ,0xAA ,0xAE ,0xA7 ,0x92 ,0x41 ,0x1 ,0x14 ,0xDC ,0xFF ,0xF3 ,0x9F ,0x61 ,0x52 ,0x51 ,0x62 ,0x94 ,0xEE ,0xFF ,0xB3 ,0x5C ,0x55 ,0x55 ,0x55 ,0x55 ,0x7A ,0xE9 ,0xFF ,0xFF ,0xD5 ,0xC9 ,0xAF ,0x18 ,0x0 ,0x2 ,0x7C ,0xF3 ,0xFF ,0xF7 ,0xF8 ,0xFE ,0xEC ,0x4F ,0x0 ,0x3E ,0xE4 ,0xFD ,0xE8 ,0xD9 ,0xF5 ,0xFD ,0xCA ,0x17 ,0x15 ,0xDC ,0xFE ,0x9E ,0x18 ,0x1 ,0x13 ,0x19 ,0x1 ,0xE ,0x8C ,0xFF ,0x91 ,0xF ,0x1 ,0x0 ,0x0 ,0x3 ,0x3D ,0xE0 ,0xFF ,0xFF ,0xF0 ,0xD4 ,0xA2 ,0x12 ,0x0 ,0x9 ,0xA5 ,0xFE ,0xED ,0xC1 ,0xC5 ,0xF3 ,0xFF ,0x77 ,0x1 ,0x59 ,0xF5 ,0xF5 ,0xB6 ,0x9A ,0xC9 ,0xE9 ,0xD2 ,0x1F ,0x15 ,0xDC ,0xFA ,0x79 ,0x6 ,0x11 ,0x4C ,0x56 ,0x14 ,0x0 ,0x61 ,0xFE ,0xAA ,0x42 ,0x1B ,0x0 ,0x4 ,0x34 ,0x68 ,0xE7 ,0xFF ,0xFF ,0xF3 ,0xE8 ,0x8E ,0x8 ,0x0 ,0x24 ,0xCD ,0xFD ,0xA9 ,0x29 ,0x34 ,0xC3 ,0xFF ,0xBB ,0x6 ,0x7A ,0xF9 ,0xE4 ,0x5E ,0x20 ,0x3D ,0x5C ,0x50 ,0x10 ,0x15 ,0xDC ,0xE8 ,0x4B ,0x1 ,0x46 ,0xD6 ,0xE0 ,0x5A ,0x0 ,0x2A ,0xF8 ,0xF7 ,0xE8 ,0x73 ,0x0 ,0x10 ,0xD7 ,0xF0 ,0xFD ,0xFF ,0xFF ,0xF1 ,0xF0 ,0x82 ,0x4 ,0x0 ,0x33 ,0xD6 ,0xFE ,0xE6 ,0xCA ,0xCC ,0xED ,0xFF ,0xD2 ,0x10 ,0x62 ,0xF6 ,0xFB ,0xE0 ,0xCE ,0xC9 ,0xB7 ,0x76 ,0xA ,0x15 ,0xDC ,0xE0 ,0x42 ,0x0 ,0x13 ,0x33 ,0x34 ,0x19 ,0x0 ,0x15 ,0xEE ,0xFF ,0xFA ,0x7C ,0x0 ,0x10 ,0xE8 ,0xFF ,0xFF ,0xFF ,0xFF ,0xF1 ,0xF1 ,0x82 ,0x4 ,0x0 ,0x35 ,0xD7 ,0xFD ,0xC3 ,0x84 ,0x83 ,0x81 ,0x7F ,0x66 ,0x9 ,0xA ,0x7E ,0xBF ,0xD0 ,0xD4 ,0xE4 ,0xFB ,0xF1 ,0x60 ,0x14 ,0xDC ,0xDF ,0x41 ,0x1 ,0x36 ,0x8F ,0x93 ,0x96 ,0x97 ,0xA2 ,0xF7 ,0xFF ,0xFA ,0x7A ,0x0 ,0x10 ,0xE7 ,0xFF ,0xFF ,0xFF ,0xFF ,0xF1 ,0xDC ,0x9B ,0xE ,0x0 ,0x13 ,0xBA ,0xFE ,0xD1 ,0x72 ,0x71 ,0xCE ,0xF3 ,0xB7 ,0x9 ,0x5F ,0xD3 ,0xCE ,0x7A ,0x48 ,0x77 ,0xE6 ,0xF3 ,0x66 ,0x14 ,0xDC ,0xF5 ,0x69 ,0x4 ,0x21 ,0x82 ,0x96 ,0x3B ,0x9 ,0x2F ,0xF7 ,0xFF ,0xFA ,0x7A ,0x0 ,0x10 ,0xE7 ,0xFF ,0xFF ,0xFF ,0xFF ,0xF2 ,0xC0 ,0xB9 ,0x1E ,0x0 ,0x1 ,0x58 ,0xF2 ,0xFD ,0xF3 ,0xF3 ,0xFE ,0xF5 ,0x5F ,0x1 ,0x30 ,0xE2 ,0xFF ,0xF1 ,0xE3 ,0xED ,0xFD ,0xC3 ,0x21 ,0x14 ,0xDC ,0xFF ,0xC9 ,0x2F ,0x1 ,0xA ,0xE ,0x1 ,0xF ,0x8E ,0xFE ,0xFF ,0xFA ,0x77 ,0x0 ,0xB ,0xE6 ,0xFF ,0xFF ,0xFF ,0xFF ,0xEF ,0xB1 ,0xCA ,0x27 ,0x0 ,0x0 ,0x2D ,0xBF ,0xE0 ,0xE7 ,0xE6 ,0xE0 ,0xC3 ,0x33 ,0x0 ,0x16 ,0xAF ,0xDF ,0xE7 ,0xE8 ,0xE5 ,0xD8 ,0x8B ,0xB ,0x14 ,0xDC ,0xFF ,0xE8 ,0x69 ,0x2A ,0x20 ,0x1E ,0x26 ,0x44 ,0xC0 ,0xFF ,0xFF ,0xFB ,0x87 ,0x1F ,0x28 ,0xE9 ,0xFF ,0xFF ,0xFF ,0xFF ,0xE4 ,0x78 ,0xF3 ,0x5C ,0x1 ,0x0 ,0x0 ,0x19 ,0x40 ,0x4F ,0x4F ,0x41 ,0x1B ,0x1 ,0x0 ,0x0 ,0x15 ,0x40 ,0x4F ,0x51 ,0x4B ,0x32 ,0x7 ,0x0 ,0x14 ,0xDC ,0xFF ,0xFF ,0xF7 ,0xD3 ,0xBF ,0xBC ,0xCC ,0xEA ,0xFE ,0xFF ,0xFF ,0xFE ,0xDD ,0xBD ,0xC2 ,0xF9 ,0xFF ,0xFF ,0xFF ,0xFF ,0xAC ,0x38 ,0xE2 ,0xD2 ,0x1D ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x14 ,0xDC ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xF7 ,0x5A ,0x7 ,0x6C ,0xE9 ,0xDA ,0x6B ,0x30 ,0x1E ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x27 ,0xDD ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xF4 ,0x8D ,0x10 ,0x0 ,0xE ,0x65 ,0xD5 ,0xF3 ,0xDA ,0xBC ,0xB8 ,0xB8 ,0xB8 ,0xB8 ,0xB8 ,0xB8 ,0xB8 ,0xB8 ,0xB8 ,0xB8 ,0xB8 ,0xB8 ,0xB8 ,0xB8 ,0xB8 ,0xB8 ,0xBE ,0xF5 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFE ,0xE2 ,0x80 ,0x1A ,0x0 ,0x0 ,0x0 ,0x5 ,0x33 ,0x76 ,0xA6 ,0xB5 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB4 ,0xB5 ,0xAB ,0x84 ,0x3D ,0xC ,0x0 ,0x0};
	unsigned char esetL16[] = { 0x0 ,0x0 ,0xE ,0x3D ,0x7D ,0x98 ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9C ,0x9A ,0x85 ,0x4A ,0x18 ,0x0 ,0x0 ,0x0 ,0xC ,0x9A ,0xE5 ,0xF3 ,0xD0 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC2 ,0xE2 ,0xFE ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xED ,0xB3 ,0x1E ,0x0 ,0x15 ,0xAA ,0xE9 ,0x86 ,0x17 ,0x1 ,0x1 ,0x1 ,0x1 ,0x1 ,0x1 ,0x1 ,0x1 ,0x1 ,0x1 ,0x1 ,0x1 ,0x1 ,0x1 ,0x1 ,0x5 ,0x85 ,0xFD ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xCA ,0x28 ,0x53 ,0xF1 ,0x77 ,0x9 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x5 ,0x89 ,0xFD ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFB ,0x7C ,0xA5 ,0xD8 ,0x7 ,0x0 ,0x4 ,0x31 ,0x75 ,0x8D ,0x8B ,0x72 ,0x2A ,0x1 ,0x1 ,0x23 ,0x6F ,0x8B ,0x92 ,0x88 ,0x5F ,0x15 ,0x5 ,0x89 ,0xFD ,0xF9 ,0xBF ,0x80 ,0x6C ,0x71 ,0x92 ,0xDF ,0xFE ,0xD1 ,0x79 ,0x71 ,0x71 ,0x71 ,0x73 ,0xCA ,0xFF ,0xFF ,0xD3 ,0xCD ,0x99 ,0x2 ,0x0 ,0x31 ,0xCF ,0xFD ,0xF8 ,0xFA ,0xFD ,0xBB ,0x21 ,0x13 ,0xA4 ,0xFB ,0xED ,0xDE ,0xF8 ,0xF5 ,0x7A ,0xB ,0x89 ,0xFD ,0xBA ,0x22 ,0x1 ,0x15 ,0xD ,0x4 ,0x58 ,0xE9 ,0xAF ,0x12 ,0x1 ,0x0 ,0x1 ,0x7 ,0xA2 ,0xFF ,0xFF ,0xF2 ,0xE8 ,0x6D ,0x1 ,0x0 ,0x7A ,0xFF ,0xCC ,0x2E ,0x3C ,0xD6 ,0xF6 ,0x61 ,0x2A ,0xDD ,0xEF ,0x63 ,0x18 ,0x4E ,0x76 ,0x4C ,0xB ,0x89 ,0xF5 ,0x68 ,0x1 ,0x56 ,0xE4 ,0xBF ,0x1E ,0xF ,0xC3 ,0xF6 ,0xE1 ,0x60 ,0x0 ,0x50 ,0xDC ,0xF5 ,0xFF ,0xFF ,0xF3 ,0xF0 ,0x5E ,0x0 ,0x0 ,0x93 ,0xFF ,0xEA ,0xB8 ,0xBE ,0xED ,0xFE ,0x7A ,0x25 ,0xD0 ,0xFC ,0xDA ,0xBE ,0xB5 ,0x98 ,0x3F ,0x8 ,0x89 ,0xF1 ,0x59 ,0x0 ,0x1F ,0x45 ,0x3E ,0xB ,0x3 ,0xB2 ,0xFF ,0xFA ,0x6C ,0x0 ,0x5A ,0xF6 ,0xFF ,0xFF ,0xFF ,0xF3 ,0xF0 ,0x5D ,0x0 ,0x0 ,0x95 ,0xFF ,0xD3 ,0x77 ,0x75 ,0x72 ,0x70 ,0x36 ,0x5 ,0x45 ,0xA3 ,0xC0 ,0xC7 ,0xDF ,0xFD ,0xCE ,0x1F ,0x89 ,0xF0 ,0x59 ,0x1 ,0x4A ,0x9D ,0x9F ,0xA2 ,0xA3 ,0xE2 ,0xFF ,0xFA ,0x6B ,0x0 ,0x59 ,0xF5 ,0xFF ,0xFF ,0xFF ,0xF3 ,0xEA ,0x6A ,0x0 ,0x0 ,0x80 ,0xFF ,0xC2 ,0x14 ,0x13 ,0x78 ,0x99 ,0x48 ,0x15 ,0x68 ,0x75 ,0x3D ,0x1C ,0x65 ,0xF1 ,0xDD ,0x24 ,0x88 ,0xF4 ,0x67 ,0x1 ,0x5D ,0xF2 ,0xE3 ,0x74 ,0x62 ,0xD0 ,0xFF ,0xFA ,0x6A ,0x0 ,0x59 ,0xF5 ,0xFF ,0xFF ,0xFF ,0xF3 ,0xD0 ,0x91 ,0x1 ,0x0 ,0x3B ,0xE0 ,0xFC ,0xE1 ,0xDF ,0xFD ,0xE5 ,0x49 ,0x1B ,0xBD ,0xFE ,0xE1 ,0xB5 ,0xD4 ,0xF9 ,0xA8 ,0x16 ,0x89 ,0xFD ,0xB6 ,0x1E ,0x3 ,0x21 ,0x18 ,0x2 ,0x2E ,0xD5 ,0xFF ,0xFA ,0x69 ,0x0 ,0x57 ,0xF5 ,0xFF ,0xFF ,0xFF ,0xF2 ,0xAC ,0xCF ,0x4 ,0x0 ,0x8 ,0x4A ,0xA3 ,0xBD ,0xBE ,0xA7 ,0x55 ,0xA ,0x3 ,0x3C ,0x9E ,0xBE ,0xC1 ,0xB7 ,0x83 ,0x26 ,0x7 ,0x89 ,0xFD ,0xF7 ,0xB5 ,0x6C ,0x52 ,0x55 ,0x76 ,0xC6 ,0xFA ,0xFF ,0xFB ,0x99 ,0x52 ,0x8C ,0xF8 ,0xFF ,0xFF ,0xFF ,0xD9 ,0x5D ,0xF5 ,0x63 ,0x6 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x5 ,0x89 ,0xFD ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFD ,0x88 ,0x1C ,0xBC ,0xE0 ,0x6D ,0x8 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x4 ,0x84 ,0xFD ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xD7 ,0x32 ,0x0 ,0x19 ,0xB1 ,0xF0 ,0xDE ,0xB0 ,0xA0 ,0xA0 ,0xA0 ,0xA0 ,0xA0 ,0xA0 ,0xA0 ,0xA0 ,0xA0 ,0xA0 ,0xA0 ,0xA0 ,0xA0 ,0xA0 ,0xA2 ,0xD3 ,0xFE ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xF6 ,0xC7 ,0x30 ,0x0 ,0x0 ,0x0 ,0x16 ,0x58 ,0xA3 ,0xBA ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xBC ,0xAB ,0x68 ,0x22 ,0x0 ,0x0};
	unsigned char esetL14[] = { 0x0 ,0x1 ,0x22 ,0x70 ,0x9F ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA8 ,0xA2 ,0x7B ,0x2F ,0x3 ,0x0 ,0x2 ,0x53 ,0xD4 ,0xD3 ,0xA2 ,0x8B ,0x89 ,0x89 ,0x89 ,0x89 ,0x89 ,0x89 ,0x89 ,0x89 ,0x89 ,0x89 ,0x89 ,0x89 ,0x9E ,0xFD ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFD ,0xE1 ,0x6D ,0x6 ,0x32 ,0xDA ,0xA4 ,0x21 ,0x2 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x2E ,0xFA ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xEE ,0x4D ,0x8C ,0xC7 ,0x16 ,0x0 ,0x1 ,0x12 ,0x29 ,0x28 ,0x12 ,0x0 ,0x0 ,0x0 ,0xD ,0x28 ,0x2E ,0x24 ,0x8 ,0x0 ,0x30 ,0xFA ,0xFF ,0xFA ,0xDE ,0xCF ,0xDC ,0xF8 ,0xFF ,0xF6 ,0xD7 ,0xD4 ,0xD4 ,0xD4 ,0xDE ,0xFE ,0xFF ,0xB6 ,0xC9 ,0x79 ,0x0 ,0x1 ,0x72 ,0xE2 ,0xEE ,0xEE ,0xE0 ,0x62 ,0x1 ,0x42 ,0xD9 ,0xE9 ,0xE1 ,0xEC ,0xC5 ,0x1C ,0x30 ,0xFA ,0xE0 ,0x4B ,0x10 ,0x1A ,0x12 ,0x3C ,0xE4 ,0xCD ,0x20 ,0xF ,0xE ,0xF ,0x46 ,0xF9 ,0xFF ,0xEE ,0xE5 ,0x49 ,0x0 ,0x17 ,0xDE ,0xF1 ,0x65 ,0x71 ,0xEC ,0xDB ,0x9 ,0x92 ,0xFE ,0x8C ,0x42 ,0x87 ,0xA6 ,0x2C ,0x30 ,0xFA ,0x94 ,0x4 ,0x50 ,0xC1 ,0x5D ,0x2 ,0x90 ,0xED ,0xAB ,0x3B ,0x2 ,0x6F ,0xBD ,0xFD ,0xFF ,0xF5 ,0xF0 ,0x36 ,0x0 ,0x30 ,0xEB ,0xF4 ,0xA7 ,0xB0 ,0xF1 ,0xFB ,0x17 ,0x8D ,0xFF ,0xD7 ,0xAF ,0xA1 ,0x77 ,0x13 ,0x30 ,0xFA ,0x75 ,0x1 ,0x30 ,0x58 ,0x36 ,0x0 ,0x6C ,0xFC ,0xF8 ,0x5A ,0x3 ,0xA9 ,0xFF ,0xFF ,0xFF ,0xF4 ,0xF0 ,0x35 ,0x0 ,0x32 ,0xEC ,0xEA ,0x69 ,0x66 ,0x64 ,0x61 ,0xA ,0x1B ,0x7F ,0xAF ,0xB8 ,0xDF ,0xFF ,0x7D ,0x30 ,0xFA ,0x74 ,0x2 ,0x61 ,0xAB ,0xAC ,0xAD ,0xCF ,0xFE ,0xF8 ,0x59 ,0x3 ,0xA7 ,0xFF ,0xFF ,0xFF ,0xF4 ,0xE7 ,0x46 ,0x0 ,0x1B ,0xE2 ,0xEE ,0x4B ,0x49 ,0xB3 ,0xBE ,0xF ,0x63 ,0xA8 ,0x6C ,0x33 ,0x8A ,0xFF ,0x85 ,0x2F ,0xFA ,0x92 ,0x4 ,0x54 ,0xCD ,0x84 ,0x34 ,0x93 ,0xFD ,0xF8 ,0x58 ,0x3 ,0xA7 ,0xFF ,0xFF ,0xFF ,0xF4 ,0xCD ,0x71 ,0x0 ,0x2 ,0x8A ,0xEE ,0xE9 ,0xE8 ,0xF0 ,0xA2 ,0x5 ,0x59 ,0xEB ,0xEA ,0xD3 ,0xE5 ,0xDD ,0x3B ,0x30 ,0xFA ,0xDD ,0x43 ,0xF ,0x1D ,0x12 ,0x23 ,0xC4 ,0xFF ,0xF8 ,0x5D ,0xB ,0xA9 ,0xFF ,0xFF ,0xFF ,0xF0 ,0x95 ,0xC0 ,0xF ,0x0 ,0x5 ,0x2D ,0x4E ,0x4E ,0x35 ,0x7 ,0x0 ,0x3 ,0x28 ,0x4E ,0x51 ,0x46 ,0x16 ,0x0 ,0x30 ,0xFA ,0xFF ,0xF5 ,0xCD ,0xBB ,0xC7 ,0xEC ,0xFF ,0xFF ,0xFD ,0xD4 ,0xBE ,0xE8 ,0xFF ,0xFF ,0xFF ,0xBF ,0x3A ,0xDF ,0x92 ,0x15 ,0x1 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x2D ,0xFA ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xF3 ,0x59 ,0x4 ,0x66 ,0xDC ,0xC5 ,0x88 ,0x6D ,0x6C ,0x6C ,0x6C ,0x6C ,0x6C ,0x6C ,0x6C ,0x6C ,0x6C ,0x6C ,0x6C ,0x6C ,0x85 ,0xFC ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xEC ,0x81 ,0x9 ,0x0 ,0x3 ,0x31 ,0x92 ,0xBE ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC5 ,0xC1 ,0x9F ,0x3F ,0x6 ,0x0};
	unsigned char esetL12[] = { 0x0 ,0x0 ,0x49 ,0xA5 ,0xBA ,0xBA ,0xBA ,0xBA ,0xBA ,0xBA ,0xBA ,0xBA ,0xBA ,0xBA ,0xBA ,0xBA ,0xB5 ,0xB4 ,0xB5 ,0xB5 ,0xB5 ,0xB5 ,0xB5 ,0xB5 ,0xB5 ,0xB5 ,0xB5 ,0xA9 ,0x5C ,0x0 ,0x0 ,0x0 ,0xB0 ,0xFF ,0x7F ,0x4C ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4B ,0x4A ,0xF4 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFE ,0xFF ,0xD4 ,0x1 ,0x65 ,0xEF ,0x1 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0xF1 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFE ,0xFE ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0x8F ,0xC3 ,0x60 ,0x0 ,0x0 ,0x7B ,0xAD ,0xAE ,0x7C ,0x0 ,0x0 ,0x60 ,0xAB ,0xB8 ,0xA4 ,0x37 ,0x0 ,0xF1 ,0xFF ,0xA3 ,0x53 ,0x46 ,0x65 ,0xF5 ,0xFF ,0x53 ,0x4F ,0x4F ,0x49 ,0xD9 ,0xFF ,0xE4 ,0xF1 ,0x11 ,0x0 ,0x6B ,0xFF ,0x9D ,0x96 ,0xFF ,0x82 ,0x14 ,0xFF ,0xC1 ,0x5C ,0xDA ,0xE7 ,0x0 ,0xF1 ,0xEE ,0x0 ,0x4E ,0x9C ,0x6 ,0x31 ,0xFF ,0x73 ,0x13 ,0x0 ,0x65 ,0xE1 ,0xFF ,0xF6 ,0xFD ,0x0 ,0x0 ,0xA7 ,0xFF ,0x94 ,0x9F ,0xFF ,0xD5 ,0x14 ,0xFF ,0xE6 ,0xA0 ,0x83 ,0x42 ,0x0 ,0xF1 ,0xB2 ,0x0 ,0x52 ,0x61 ,0x3 ,0x0 ,0xFF ,0xFF ,0x30 ,0x0 ,0xFF ,0xFF ,0xFF ,0xF5 ,0xFD ,0x0 ,0x0 ,0xAA ,0xFF ,0x48 ,0x50 ,0x40 ,0x34 ,0x0 ,0x4C ,0x97 ,0xAE ,0xF1 ,0xFF ,0x0 ,0xF1 ,0xB0 ,0x0 ,0xA9 ,0xBD ,0xCC ,0xCC ,0xFE ,0xFF ,0x2F ,0x0 ,0xFC ,0xFF ,0xFF ,0xF5 ,0xF3 ,0xD ,0x0 ,0x75 ,0xFF ,0x85 ,0x71 ,0xFF ,0xC8 ,0x14 ,0xE9 ,0xB9 ,0x37 ,0xB7 ,0xFF ,0x0 ,0xF1 ,0xED ,0x0 ,0x4B ,0xA5 ,0x5 ,0x0 ,0xFF ,0xFF ,0x2E ,0x0 ,0xFC ,0xFF ,0xFF ,0xF6 ,0xC9 ,0x56 ,0x0 ,0x6 ,0xA0 ,0xD9 ,0xDB ,0xB0 ,0xD ,0x0 ,0x8F ,0xD6 ,0xE2 ,0xCC ,0x60 ,0x0 ,0xF1 ,0xFF ,0x98 ,0x3F ,0x2E ,0x4E ,0xD8 ,0xFF ,0xFF ,0x59 ,0x2A ,0xFC ,0xFF ,0xFF ,0xE8 ,0x71 ,0xE2 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0x0 ,0xF1 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFE ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0x9A ,0x0 ,0xCB ,0xEE ,0x61 ,0x30 ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0x2F ,0xF2 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFE ,0xFF ,0xE8 ,0x7 ,0x0 ,0x0 ,0x65 ,0xC2 ,0xD6 ,0xD5 ,0xD5 ,0xD5 ,0xD5 ,0xD5 ,0xD5 ,0xD5 ,0xD5 ,0xD5 ,0xD5 ,0xD5 ,0xCD ,0xCC ,0xCD ,0xCD ,0xCD ,0xCD ,0xCD ,0xCD ,0xCD ,0xCD ,0xCD ,0xC4 ,0x78 ,0x0 ,0x0};
	unsigned char esetL10[] = { 0x1 ,0x30 ,0x98 ,0xB2 ,0xB2 ,0xB2 ,0xB2 ,0xB2 ,0xB2 ,0xB2 ,0xB2 ,0xB2 ,0xB2 ,0xBF ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC1 ,0xC0 ,0xA1 ,0x3C ,0x2 ,0x39 ,0xD1 ,0x66 ,0x2D ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x29 ,0x30 ,0xE6 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xF1 ,0x4D ,0xA9 ,0x55 ,0x1 ,0x15 ,0x36 ,0x2F ,0xA ,0x0 ,0x15 ,0x37 ,0x37 ,0x15 ,0xA ,0xE3 ,0xF9 ,0xD4 ,0xC4 ,0xE1 ,0xFE ,0xDA ,0xC7 ,0xC7 ,0xD3 ,0xFF ,0xC9 ,0xC5 ,0x15 ,0x1D ,0xDF ,0xCE ,0xE1 ,0xA7 ,0x15 ,0xD5 ,0xC5 ,0xC7 ,0xC0 ,0x14 ,0xE2 ,0x7F ,0x18 ,0x45 ,0x13 ,0xCB ,0x7C ,0x14 ,0xF ,0x65 ,0xFF ,0xF6 ,0xC8 ,0x5 ,0x49 ,0xFD ,0x91 ,0xB8 ,0xF1 ,0x2B ,0xED ,0xAE ,0x87 ,0x69 ,0xF ,0xDD ,0x3C ,0x32 ,0x79 ,0x11 ,0x8E ,0xFB ,0x5A ,0x42 ,0xF6 ,0xFF ,0xF7 ,0xC8 ,0x5 ,0x4B ,0xFC ,0x61 ,0x59 ,0x68 ,0xB ,0x64 ,0x8F ,0xAB ,0xF3 ,0x3B ,0xDD ,0x3C ,0x4F ,0xC0 ,0xA1 ,0xD1 ,0xFF ,0x5A ,0x43 ,0xFB ,0xFF ,0xF7 ,0xC5 ,0x12 ,0x22 ,0xEA ,0xBD ,0xCF ,0xCC ,0x1B ,0xD0 ,0xB7 ,0xAA ,0xE4 ,0x2B ,0xE2 ,0x7A ,0x18 ,0x4F ,0x11 ,0xB3 ,0xFF ,0x5A ,0x42 ,0xFB ,0xFF ,0xF7 ,0xAE ,0x4C ,0x1 ,0x25 ,0x55 ,0x50 ,0x19 ,0x1 ,0x29 ,0x57 ,0x55 ,0x23 ,0xA ,0xE3 ,0xF8 ,0xC8 ,0xB3 ,0xD1 ,0xFC ,0xFF ,0xCD ,0xC6 ,0xFE ,0xFF ,0xD0 ,0x42 ,0xCF ,0x53 ,0x20 ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x1C ,0x23 ,0xE5 ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xFF ,0xF6 ,0x58 ,0x2 ,0x3D ,0xAE ,0xC0 ,0xBE ,0xBE ,0xBE ,0xBE ,0xBE ,0xBE ,0xBE ,0xBE ,0xBF ,0xD2 ,0xD5 ,0xD5 ,0xD5 ,0xD5 ,0xD5 ,0xD5 ,0xD5 ,0xD5 ,0xBB ,0x4B ,0x3};
	std::map<int, std::pair<unsigned char *, int> > logoMap;
	logoMap[10] = std::make_pair(esetL10, 25);
	logoMap[12] = std::make_pair(esetL12, 31);
	logoMap[14] = std::make_pair(esetL14, 36);
	logoMap[16] = std::make_pair(esetL16, 41);
	logoMap[18] = std::make_pair(esetL18, 46);
	logoMap[20] = std::make_pair(esetL20, 51);
	logoMap[24] = std::make_pair(esetL24, 61);

	if (logoMap.find(fontY) != logoMap.end())
	{
		int logoX = logoMap.find(fontY)->second.second;
		unsigned char *logo = logoMap.find(fontY)->second.first;

		// in case the font is too narrow, add some spaces
		int numEsetChars = 5;
		while (fontX * numEsetChars<logoX + 1)
		{
			std::wcout << L" ";
			numEsetChars++;
		}
		std::wcout << " " << welcomeMessage << " " << wchVersionString << std::endl;

		HWND myconsole = GetConsoleWindow();
		HDC mydc = GetDC(myconsole);
		for (int x = 0; x < std::max<int>(logoX, fontX * numEsetChars); ++x)
		for (int y = 0; y < fontY; ++y)
		{
			unsigned char col = 0;
			if (x<logoX)
			{
				col = logo[x + y * logoX];
			}
			SetPixel(mydc, 1 + fontX * cursorX + x, fontY * cursorY + y, RGB(col, col, col));
		}
		ReleaseDC(myconsole, mydc);
	}
	else std::wcout << " " << welcomeMessage << " " << wchVersionString << std::endl;

	RunStartupScript(scriptInterpreter, partiallyFailed);

	ConsoleKeyboardInput keyInput(scriptInterpreter);

	std::wcout << std::endl << gConsoleColors.LIGHT_YELLOW << L"> " << gConsoleColors.DEFAULT;
	while (1)
	{
		keyInput.GetInput();
		std::wcout << std::endl;
		bool commandExecuted = false;
		partiallyFailed = false;
		/*DWORD result = */ExecuteCommands(scriptInterpreter, keyInput.GetLine(), false, commandExecuted, partiallyFailed);
		if (scriptInterpreter.ShouldExit())
			break;

		if (commandExecuted)
			std::wcout << std::endl;

		std::wcout << gConsoleColors.LIGHT_YELLOW << L"> " << gConsoleColors.DEFAULT;

	}

	if (gCmdHistory.Active)
	{
		gCmdHistory.Save();
	}
	else
	{
		if (FileExists(histFile.c_str()))
			DeleteFile(histFile.c_str());
	}
}

void RunStartupScript(ScriptInterpreter &scriptInterpreter, bool &partiallyFailed) {
	std::wstring startupScriptFile = scriptInterpreter.GetConnectionParams().alternativeStartupScript;
	if (startupScriptFile.empty()) {
		startupScriptFile = GetProgramDataFilePath(L"eracmd_startup.txt");

		if (! FileExists(startupScriptFile.c_str()))
			return;
	}

	FileLoader fileLoader(startupScriptFile);
	std::wstring startupScript;
	try
	{
		startupScript = fileLoader.Load();
	}
	catch (const EraCMDException &e)
	{
		e.Print(std::wcerr);
		return;
	}

	bool commandExecuted = false;
	partiallyFailed = false;
	/*DWORD result = */ExecuteCommands(scriptInterpreter, startupScript , false, commandExecuted, partiallyFailed);
	scriptInterpreter.ResetExit();

	if (commandExecuted)
		std::wcout << std::endl;
}

