/**
 * @file FlagValuesConversions.h
 * @author Copyright (c) 2014 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "FlagValuesConversions.h"

#include <Common/era_api/err_codes.h>
#include "../EracmdExceptions.h"

namespace FlagValuesConversions {

std::wstring FromBool(const bool value)
{
	return value ? L"true" : L"false";
}

bool ToBool(const std::wstring &str)
{
	if (str.empty())
	{
		return true;
	}

	if (::_wcsicmp(str.c_str(), L"true") == 0)
	{
		return true;
	}

	if (::_wcsicmp(str.c_str(), L"false") == 0)
	{
		return false;
	}

	throw EraCMDException(CMDERROR_INVALID_BOOL_ARGUMENT, str.c_str(), L"true", L"false");
}

std::wstring FromBoolWithCustomNames(const bool value, const std::wstring &trueValue, const std::wstring &falseValue)
{
	return value ? trueValue : falseValue;
}

bool ToBoolWithCustomNames(const std::wstring &str, const std::wstring &trueValue, const std::wstring &falseValue)
{
	if (::_wcsicmp(str.c_str(), trueValue.c_str()) == 0)
	{
		return true;
	}

	if (::_wcsicmp(str.c_str(), falseValue.c_str()) == 0)
	{
		return false;
	}

	throw EraCMDException(CMDERROR_INVALID_BOOL_ARGUMENT, str.c_str(), trueValue.c_str(), falseValue.c_str());
}

std::wstring FromOutputFormatting(const CommandFlags::OuputFormatting value)
{
	switch (value)
	{
		case CommandFlags::FORMAT_CSV:
			return L"csv";
		case CommandFlags::FORMAT_TABLE:
			return L"table";
		default:
			throw std::exception("Unsupported value for OutputFormatting");
	}
}

CommandFlags::OuputFormatting ToOutputFormatting(const std::wstring &str)
{
	if (::_wcsicmp(str.c_str(), L"csv") == 0)
	{
		return CommandFlags::FORMAT_CSV;
	}
	if (::_wcsicmp(str.c_str(), L"table") == 0)
	{
		return CommandFlags::FORMAT_TABLE;
	}

	throw EraCMDException(CMDERROR_INVALID_FORMATTING, str.c_str());
}

std::wstring FromEncoding(const CommandFlags::Encoding value)
{
	switch (value)
	{
		case CommandFlags::ENCODING_ANSI:
			return L"ansi";
		case CommandFlags::ENCODING_UTF8:
			return L"utf8";
		case CommandFlags::ENCODING_UTF16:
			return L"utf16";
		default:
			throw std::exception("Unsupported value for Encoding");
	}
}

CommandFlags::Encoding ToEncoding(const std::wstring &str)
{
	if (::_wcsicmp(str.c_str(), L"ansi") == 0)
	{
		return CommandFlags::ENCODING_ANSI;
	}
	if (::_wcsicmp(str.c_str(), L"utf8") == 0)
	{
		return CommandFlags::ENCODING_UTF8;
	}
	if (::_wcsicmp(str.c_str(), L"utf16") == 0)
	{
		return CommandFlags::ENCODING_UTF16;
	}

	throw EraCMDException(CMDERROR_INVALID_ENCODING, str.c_str());
}

std::wstring FromOutputMode(const CommandFlags::OutputMode value)
{
	switch (value)
	{
		case CommandFlags::MODE_OVERWRITE:
			return L"o";
		case CommandFlags::MODE_APPEND:
			return L"a";
		default:
			throw std::exception("Unsupported value for Mode");
	}
}

CommandFlags::OutputMode ToOutputMode(const std::wstring &str)
{
	if (::_wcsicmp(str.c_str(), L"o") == 0)
	{
		return CommandFlags::MODE_OVERWRITE;
	}
	if (::_wcsicmp(str.c_str(), L"a") == 0)
	{
		return CommandFlags::MODE_APPEND;
	}

	throw EraCMDException(CMDERROR_INVALID_MODE, str.c_str());
}

std::wstring FromHeaderType(const CommandFlags::HeaderType value)
{
	switch (value)
	{
		case CommandFlags::HEADER_KEYWORD:
			return L"keyword";
		case CommandFlags::HEADER_PRETTY:
			return L"pretty";
		case CommandFlags::HEADER_NONE:
			return L"none";
		default:
			throw std::exception("Unsupported value for HeaderType");
	}
}

CommandFlags::HeaderType ToHeaderType(const std::wstring &str)
{
	if (::_wcsicmp(str.c_str(), L"keyword") == 0)
	{
		return CommandFlags::HEADER_KEYWORD;
	}
	if (::_wcsicmp(str.c_str(), L"pretty") == 0)
	{
		return CommandFlags::HEADER_PRETTY;
	}
	if (::_wcsicmp(str.c_str(), L"none") == 0)
	{
		return CommandFlags::HEADER_NONE;
	}

	throw EraCMDException(CMDERROR_INVALID_HEADER_TYPE, str.c_str());
}

std::wstring FromFieldFormatting(const CommandFlags::FieldFormatting value)
{
	switch (value)
	{
		case CommandFlags::FIELD_KEYWORD:
			return L"keyword";
		case CommandFlags::FIELD_PRETTY:
			return L"pretty";
		default:
			throw std::exception("Unsupported value for FieldFormatting");
	}
}

CommandFlags::FieldFormatting ToFieldFormatting(const std::wstring &str)
{
	if (::_wcsicmp(str.c_str(), L"keyword") == 0)
	{
		return CommandFlags::FIELD_KEYWORD;
	}
	if (::_wcsicmp(str.c_str(), L"pretty") == 0)
	{
		return CommandFlags::FIELD_PRETTY;
	}

	throw EraCMDException(CMDERROR_INVALID_FIELD_FORMATTING, str.c_str());
}

std::wstring FromDelimiter(const wchar_t value)
{
	if (value == L'')
	{
		return std::wstring();
	}
	else
	{
		return std::wstring(1, value);
	}
}

wchar_t ToDelimiter(const std::wstring &str)
{
	if (str.length() == 0)
		return L'';
	else if (str.length() == 1)
		return str[0];
	else
		throw EraCMDException(CMDERROR_DELIM_TOO_LONG, str.c_str());
}

std::wstring FromString(const std::wstring &value)
{
	return value;
}

std::wstring ToString(const std::wstring &str)
{
	return str;
}

std::wstring ToWritablePath(const std::wstring &str)
{
	if (str.empty())
		return str;

	bool fileExists = false;

	DWORD attrs = GetFileAttributes(str.c_str());
	if (attrs != INVALID_FILE_ATTRIBUTES)
	{
		if (attrs & FILE_ATTRIBUTE_DIRECTORY)
			throw EraCMDException(CMDERROR_PATH_IS_A_DIRECTORY, str.c_str());

		fileExists = true;
	}

	{
		std::ofstream os(str.c_str(), std::ios::app);
		if (! os.good())
		{
			if (fileExists)
				throw EraCMDException(CMDERROR_FILE_NOT_WRITABLE, str.c_str());
			else
				throw EraCMDException(CMDERROR_FILE_CANNOT_BE_CREATED, str.c_str());
		}
	}

	if (! fileExists)
		::DeleteFile(str.c_str());

	return str;
}

}

