/**
 * @file PagingStream.cpp
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "PagingStream.h"

#include <conio.h>

PagingStreambuf::PagingStreambuf(std::wstreambuf &dest, std::function<void()> hintPrintingFunc, std::function<void()> headerPrintingFunc)
:
	dest(dest),
	hintPrintingFunc(hintPrintingFunc),
	headerPrintingFunc(headerPrintingFunc),
	stdOutputHandle(::GetStdHandle(STD_OUTPUT_HANDLE)),
	printingHint(false),
	printingHeader(false),
	exited(false),
	disabled(false)
{
	SetInitialCursorPosition();
	PrintHeader();
}

PagingStreambuf::int_type PagingStreambuf::overflow(int_type ch)
{
	if (exited)
	{
		return 0;
	}

	if (disabled)
	{
		return dest.sputc(ch);
	}

	if (! printingHint)
	{
		CONSOLE_SCREEN_BUFFER_INFO info;
		::GetConsoleScreenBufferInfo(stdOutputHandle, &info);

		COORD currentCursorPosition = info.dwCursorPosition;

		int windowHeight = info.srWindow.Bottom - info.srWindow.Top + 1;
		bool scrollable = (windowHeight < info.dwSize.Y);

		int scrollBy = 1;

		if (currentCursorPosition.Y >= info.dwSize.Y - scrollBy)
		{
			SMALL_RECT origRect;
			origRect.Left = 0;
			origRect.Top = 0;
			origRect.Right = info.dwSize.X;
			origRect.Bottom = info.dwSize.Y;
			COORD destinationOrigin;
			destinationOrigin.X = 0;
			destinationOrigin.Y = -scrollBy;

			CHAR_INFO charInfo;
			charInfo.Char.UnicodeChar = L' ';
			charInfo.Attributes = 0;

			::ScrollConsoleScreenBuffer(stdOutputHandle, &origRect, NULL, destinationOrigin, &charInfo);

			COORD pos = info.dwCursorPosition;
			pos.Y -= scrollBy;
			::SetConsoleCursorPosition(stdOutputHandle, pos);
			initialCursorPosition -= scrollBy;
			currentCursorPosition.Y -= scrollBy;

			SMALL_RECT newWindowPos = info.srWindow;
			newWindowPos.Top -= scrollBy;
			newWindowPos.Bottom -= scrollBy;

			::SetConsoleWindowInfo(stdOutputHandle, TRUE, &newWindowPos);
		}

		if ((currentCursorPosition.Y - initialCursorPosition) >= windowHeight - 1 - (scrollable ? 0 : 1))
		{
			COORD cursorBeforeHint = currentCursorPosition;
			PrintHint();

			COORD cursorAfterHint;
			{
				CONSOLE_SCREEN_BUFFER_INFO info;
				::GetConsoleScreenBufferInfo(stdOutputHandle, &info);

				cursorAfterHint = info.dwCursorPosition;
			}

			while (true)
			{
				int c = _getch();
				if (c == 13)
				{
					break;
				}
				if ((c == 'e') || (c == 'E'))
				{
					exited = true;
					break;
				}
				if (c == 27)
				{
					disabled = true;
					break;
				}
			}

			if (cursorBeforeHint.Y == cursorAfterHint.Y)
			{
				int length = cursorAfterHint.X - cursorBeforeHint.X;

				DWORD numberOfWritten_ignored;
				::FillConsoleOutputCharacter(stdOutputHandle, L' ', length, cursorBeforeHint, &numberOfWritten_ignored);
				::FillConsoleOutputAttribute(stdOutputHandle, 0, length, cursorBeforeHint, &numberOfWritten_ignored);
			}

			::SetConsoleCursorPosition(stdOutputHandle, cursorBeforeHint);

			if (exited)
			{
				return 0;
			}

			if (disabled)
			{
				return dest.sputc(ch);
			}

			SetInitialCursorPosition();
			PrintHeader();
		}
	}

	return dest.sputc(ch);
}

PagingStreambuf::int_type PagingStreambuf::underflow()
{
	return std::wstreambuf::underflow();
};

int PagingStreambuf::sync()
{
	return std::wstreambuf::sync();
};

std::wstreambuf *PagingStreambuf::setbuf(wchar_t *p, std::streamsize len)
{
	return std::wstreambuf::setbuf(p, len);
}

void PagingStreambuf::SetInitialCursorPosition()
{
	CONSOLE_SCREEN_BUFFER_INFO info;
	::GetConsoleScreenBufferInfo(stdOutputHandle, &info);

	initialCursorPosition = info.dwCursorPosition.Y;
}

void PagingStreambuf::PrintHint()
{
	printingHint = true;

	CONSOLE_SCREEN_BUFFER_INFO info;
	bool infoValid = (::GetConsoleScreenBufferInfo(stdOutputHandle, &info) != 0);

	::SetConsoleTextAttribute(stdOutputHandle, 16 * 7);

	hintPrintingFunc();

	if (infoValid)
	{
		::SetConsoleTextAttribute(stdOutputHandle, info.wAttributes);
	}

	printingHint = false;
}

void PagingStreambuf::PrintHeader()
{
	if (!printingHeader)
	{
		printingHeader = true;

		CONSOLE_SCREEN_BUFFER_INFO info;
		bool infoValid = (::GetConsoleScreenBufferInfo(stdOutputHandle, &info) != 0);

		headerPrintingFunc();

		if (infoValid)
		{
			::SetConsoleTextAttribute(stdOutputHandle, info.wAttributes);
		}

		printingHeader = false;
	}
}

StreamModifier_Paging::StreamModifier_Paging(std::wostream &origStream, const bool enablePaging, std::function<void()> hintPrintingFunc, std::function<void()> headerPrintingFunc)
:
	origStream(origStream),
	pagingEnabled(false)
{
	if (enablePaging && (&origStream == &std::wcout) && (origStream.rdbuf() != NULL))
	{
		prevRdbuf = origStream.rdbuf(new PagingStreambuf(*origStream.rdbuf(), hintPrintingFunc, headerPrintingFunc));
		pagingEnabled = true;
	}
	else
	{
		headerPrintingFunc();
	}
}

StreamModifier_Paging::~StreamModifier_Paging()
{
	if (pagingEnabled)
	{
		delete origStream.rdbuf(prevRdbuf);
	}
}

