/**
 * @file stringUtils.cpp
 * @author Copyright (c) 2013 ESET, spol. s r.o.
 */

#include "set_nod.h"
#include "stringUtils.h"

#include <Common/era_api/err_codes.h>
#include "../EracmdExceptions.h"

std::string WstringToString(const std::wstring &str)
{
	std::string res;
	res.assign(str.begin(), str.end());
	return res;
}

// check if string contains only numerals
void StrConversionCheck(const std::wstring &str)
{
	for (size_t i = 0; i < str.length(); ++i)
	{
		if (str[i]<L'0' || str[i]>L'9')
		{
			throw std::exception("can not convert string value to number");
		}
	}
}

uint32_t WstringToUint(const std::wstring &str)
{
	StrConversionCheck(str);
	int result = std::stoi(str);
	return result;
}

int32_t WstringToInt(const std::wstring &str)
{
	StrConversionCheck(str);
	int32_t result = std::stoul(str);
	return result;
}

uint64_t WstringToUInt64(const std::wstring &str)
{
	StrConversionCheck(str);
	uint64_t result = std::stoull(str);
	return result;
}

int64_t WstringToInt64(const std::wstring &str)
{
	StrConversionCheck(str);
	int64_t result = std::stoll(str);
	return result;
}

void SplitToWords(const std::wstring &src, WCHAR wordDelimiter, std::vector<std::wstring> &outResult, bool allowEmptyWords)
{
	std::wstring tempS;
	for (size_t i = 0; i < src.length(); ++i)
	{
		if (src[i] != wordDelimiter)
			tempS += src[i];
		else
		if (allowEmptyWords || !tempS.empty())
		{
			outResult.push_back(tempS);
			tempS.clear();
		}
	}
	if (!tempS.empty())
		outResult.push_back(tempS);
}

std::wstring LoadStringFromResource(unsigned int stringId)
{
	WCHAR *roBuffer = NULL;
	HINSTANCE handle = GetModuleHandle(NULL);
	int len = LoadStringW(handle, stringId, reinterpret_cast<LPWSTR>(&roBuffer), 0);
	if (len)
		return std::wstring(roBuffer, len);
	else
		return std::wstring();
}

std::wstring LoadStringFromFileInResource(unsigned int stringId, const std::wstring &resourceType)
{
	HINSTANCE handle = GetModuleHandle(NULL);
	HRSRC hRes = FindResource(handle, MAKEINTRESOURCE(stringId), resourceType.c_str());
	DWORD dwSize = SizeofResource(handle, hRes);
	HGLOBAL hGlob = LoadResource(handle, hRes);
	const WCHAR *pData = reinterpret_cast<const WCHAR *>(::LockResource(hGlob));
	return std::wstring(pData + 1, (dwSize-2)/2);
}

int ReplaceAllSubstrings(std::wstring &str, const std::wstring &replaceFrom, const std::wstring &replaceTo)
{
	int result = 0;
	size_t offset = str.find(replaceFrom);
	while (offset != std::wstring::npos)
	{
		result++;
		str.replace(offset, replaceFrom.length(), replaceTo);
		offset += replaceTo.length();
		offset = str.find(replaceFrom, offset);
	}
	return result;
}

std::wstring ItemAndDescriptionFormatter::FormatOutput(unsigned int maxLineWidth)
{
	std::wstringstream strm;

	size_t descOffset = 0;
	for (size_t i = 0; i < data.size(); ++i)
		descOffset = std::max<size_t>(descOffset, data[i].first.length());

	descOffset += 4; // two spaces at the beginning, two in the middle

	for (size_t i = 0; i < data.size(); ++i)
	{
		std::vector<std::wstring> desc;
		std::wstring item = data[i].first;
		if (maxLineWidth <= 0)
			desc.push_back(data[i].second);
		else
			WrapString(data[i].second, maxLineWidth - descOffset, desc);

		size_t descSize = desc.size();
		strm << L"  " << item << std::wstring(descOffset - item.length()-2, L' ') << (descSize > 0 ? desc[0] : L"") << std::endl;
		for (size_t j = 1; j < descSize; ++j)
		{
			strm << std::wstring(descOffset, L' ') << desc[j] << std::endl;
		}
	}

	return strm.str();
}

void ItemAndDescriptionFormatter::SortByItem()
{
	std::sort(data.begin(), data.end(), [](const std::pair<std::wstring, std::wstring>& dataItem1, const std::pair<std::wstring, std::wstring>& dataItem2) -> bool
	{
		return dataItem1.first < dataItem2.first;
	});
}

std::wstring GetFirstSubstringBeforeDelimiter(std::wstring &src, const std::wstring &delimiter, bool removeFromSrc)
{
	std::wstring subString;
	size_t delPos = src.find(delimiter);
	if (delPos == std::wstring::npos)
	{
		subString = src;
		if (removeFromSrc)
			src.clear();
	}
	else
	{
		subString = src.substr(0, delPos);
		if (removeFromSrc)
			src.erase(0, delPos + delimiter.size());
	}
	return subString;
}

// the source should not contain \t. this function takes whitespaces as one char (so space only should be used)
void WrapString(const std::wstring &src, size_t screenWidth, std::vector<std::wstring> &outResult)
{
	std::wstring line;
	size_t pos = 0, lineStart = 0, wordStart = -1;
	while (pos<src.length())
	{
		WCHAR ch = src[pos];
		bool endLine = false;
		if (ch == L'\n')
		{
			endLine = true; // endline
			pos++;
		}
		else
		if (ch <= ' ') // whitespace
		{
			wordStart = -1;
			pos++;
			if (line.length() < screenWidth) // we can skip whitespaces if it is at the end of line
				line += ch;
		}
		else // letter
		{
			if (wordStart == -1)
				wordStart = pos;
			if (line.length() >= screenWidth)	// too long, rollback last word
			{
				endLine = true;
				size_t lenToDelete = pos - wordStart;
				if (lenToDelete>=screenWidth) // word too long for single line, do not rollback
					pos--;
				else
				{
					line.erase(line.length() - lenToDelete, lenToDelete);
					pos = wordStart;
					wordStart = -1;
				}
			}
			else
			{
				line += ch;
				pos++;
			}
		}

		if (endLine)
		{
			outResult.push_back(line);
			line.clear();
			lineStart = pos;
			wordStart = -1;
		}
	}
	if (!line.empty())
		outResult.push_back(line);
}

std::wstring WrapString(const std::wstring &str, size_t screenWidth, bool addNewLineEvenIfSizeMatches)
{
	std::vector<std::wstring> data;
	WrapString(str, screenWidth, data);
	std::wstring result;
	for (size_t i = 0; i < data.size(); ++i)
	{
		result += data[i];
		if (data[i].length()<screenWidth || (data[i].length() == screenWidth && addNewLineEvenIfSizeMatches))
			result += L"\n";
	}
	return result;
}

FILETIME ToFileTime(const std::wstring &str)
{
	if (str.size() < 10)
	{
		throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());
	}

	int year = 0; try { year = WstringToUint(str.substr(0, 4)); } catch (...) { throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str()); }
	if (! (year >= 0)) throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());

	if (str.substr(4, 1) != L"-") throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());

	int month = 0; try { month = WstringToUint(str.substr(5, 2)); } catch (...) { throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str()); }
	if (! ((month >= 1) && (month <= 12))) throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());

	if (str.substr(7, 1) != L"-") throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());

	int day = 0; try { day = WstringToUint(str.substr(8, 2)); } catch (...) { throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str()); }
	if (! ((day >= 1) && (day <= 31))) throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());

	int hour = 0;
	int minute = 0;
	int second = 0;
	if (str.size() > 10)
	{
		if (str.size() < 13)
		{
			throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());
		}

		if (str.substr(10, 1) != L" ") throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());

		try { hour = WstringToUint(str.substr(11, 2)); } catch (...) { throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str()); }
		if (! ((hour >= 0) && (hour <= 23))) throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());

		if (str.size() > 13)
		{
			if (str.size() < 16)
			{
				throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());
			}

			if (str.substr(13, 1) != L":") throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());

			try { minute = WstringToUint(str.substr(14, 2)); } catch (...) { throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str()); }
			if (! ((minute >= 0) && (minute <= 59))) throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());

			if (str.size() > 16)
			{
				if (str.size() < 19)
				{
					throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());
				}

				if (str.substr(16, 1) != L":") throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());

				try { second = WstringToUint(str.substr(17, 2)); } catch (...) { throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str()); }
				if (! ((second >= 0) && (second <= 59))) throw EraCMDException(CMDERROR_INVALID_DATETIME, str.c_str());
			}
		}
	}

	SYSTEMTIME st;
	::memset(&st, 0, sizeof(st));
	st.wYear = year;
	st.wMonth = month;
	st.wDay = day;
	st.wHour = hour;
	st.wMinute = minute;
	st.wSecond = second;

	FILETIME fileTime;
	::memset(&fileTime, 0, sizeof(fileTime));

	::SystemTimeToFileTime(&st, &fileTime);

	return fileTime;
}

bool ToBool(const std::wstring &str)
{
	if (::_wcsicmp(str.c_str(), L"true") == 0)
	{
		return true;
	}
	if (::_wcsicmp(str.c_str(), L"false") == 0)
	{
		return false;
	}

	throw EraCMDException(CMDERROR_INVALID_BOOL, str.c_str());
}

unsigned int ToPositiveNumber(const std::wstring &str)
{
	unsigned int number = 0;
	try
	{
		number = WstringToUint(str);
	}
	catch (...)
	{
		throw EraCMDException(CMDERROR_POSITIVE_NUMBER_EXPECTED, str.c_str());
	}

	if (number <= 0)
	{
		throw EraCMDException(CMDERROR_POSITIVE_NUMBER_EXPECTED, str.c_str());
	}

	return number;
}

int ParseHost(const std::wstring &host, std::wstring &server, uint16_t &port)
{
	std::wstring::size_type colPos = host.find(L':');
	if (colPos == std::wstring::npos)
		server = host;
	else
	{
		if (colPos > 0)
			server = host.substr(0, colPos);
		if (colPos < host.size() - 1)
		{
			std::wstring portStr = host.substr(colPos + 1);
			try
			{
				port = WstringToUint(portStr);
			}
			catch(...)
			{
				return -1;
			}
		}
	}

	return 0;
}
